'*******************************************************************************
'*  Name    : Dev_18F65K40_18F67K40_LSM_V11.BAS                                *
'*  Author  : [Alberto Freixanet]                                              *
'*  Notice  : Copyright (c) 2016 [AFreixanet]                                  *
'*          : All Rights Reserved                                              *
'*  Date    : 14/11/2017                                                       *
'*  Version : 5.0 K40 series. PIC18F27K40 (Bootloader in Low Side of Memory)   *
'*          : An Universal PIC18 Bootloader for all PIC18s                     *
'*  Notes   : TEMPLATE for PDS Bootloader PIC18 128KB ROM Memory               *
'*          : For PIC18F27K40/PIC18F47K00                                      *
'*          : The HEX file can be send by a simple COM terminal with software  *
'*          : HandShake Control XON/XOFF                                       *
'*          : For use with 18F devices and the Proton Basic V3.6.0.3 onwards   *
'*          : of Crownhill Associates Ltd                                     *
'*          : This BAS file COULD be used as a template for other PIC18FxxK40  *
'*******************************************************************************
'
(*
New:
    Version 1.0 K40 series, 31/01/2017
    Modifications of Code:
    Option _EnableWriteIDLOCS_
    Normaly the MSPAM wrote 8 bytes only.
    Now this new PIC can write 16 Bytes in ID Locations (8 Words) See the Datasheet
    The limit of number of ID was a constant = 7, now it is a variables depending of 
    the number of bytes received, (NbDataBytesAvailabled) To allow To read/write
    8 Or 16 bytes according of the Hex file.
    
    PPS configuration:
    Only 2 pins are configured by PPS For the UART only. And this pins configuration 
    is not erased after downloading the user Code.
    
    Two pins definitions have been added:
    $define OPEN_DRAINbit ODCONCBits_ODCC6  ' Define the Open Drain pin for TX1 Bit.
    $define SLEW_RATEbit SLRCONCBits_SLRC6  ' Define the Slew Rate pin for TX1 Bit.

    ROM memory has a new System To read/write.
    Some bits of the NVMCON1 must be Set before To use the Tblrd*+,Tblrd*-,tblwt*+.
    The NVMREG<1:0> control bits determine If the Access will be To Data EEPROM
    Memory locations. PFM locations Or User IDs, Configuration bits, Rev ID And Device ID.
    When NVMREG<1:0> = 00, any subsequent operations will operate On the Data EEPROM Memory.
    When NVMREG<1:0> = 10, any subsequent operations will operate On the program memory.
    When NVMREG<1:0> = x1, any subsequent operations will operate On the
    Configuration bits, User IDs, Rev ID And Device ID.

    Version V1.1 26/09/2017  
    To write the Bootloader To the High Side of ROM Memory with password control. (HSM)
    With the PDS Version 3603 the Code standard (without option is larger??)
    Added option TIMEOUT 2 And 5 sec To choose.
    Added option with PLL Or without PLL.
    Some options are moved To be more easy To read.

    Version V1.1 26/09/2017  
    To write the Bootloader To the High Side of ROM Memory with password control. (HSM)
    With the PDS Version 3603 the Code standard (without option is larger??)
    Added option TIMEOUT 2 And 5 sec To choose.
    Added option with PLL Or without PLL.
    Some options are moved To be more easy To read.
*)
'-------------------------------------------------------------------------------
$define BootloaderVersion "V1.1"
$define YES 1
$define NO 0
$define UART1 0
$define UART 0
$define UART2 1
$define UART3 2
$define UART4 3
$define UART5 4
'-------------------------------------------------------------------------------
' Version of the Bootloader to write in the User ID Locations.
$define _IDLOC0_ "L"
$define _IDLOC1_ "O"
$define _IDLOC2_ "A"
$define _IDLOC3_ "D"
$define _IDLOC4_ "E"
$define _IDLOC5_ "R"
$define _IDLOC6_ "5"
$define _IDLOC7_ "0"
'===============================================================================
(*       
 PDS PIC Bootloader template For the Devices:
    PIC18F65K40
    PIC18F66K40
    PIC18F67K40
*)                                  
'*******************************************************************************
(*
 DESCRIPTION:
    Version 2.0: Available To write upper than 64KB ROM memory. (HSM)
    Version 2.2: Available To use the internal oscillator For downloading. (HSM)
    Version 3.0: Bootloader in High Side of Memory with Random acces Code (Password). (HSM)
    Version 4.0: Bootloader in Low  Side of Memory with Random acces Code (Password). (LSM)
    Version 4.1: Rewrite the configuration section and some configuration code for bootloader with Random acces Code (Password). (LSM)
    Version 1.0 K40: Bootloader in Low & High Side of Memory with Random acces Code (Password). (LSM & HSM)
    
    This is a PDS Bootloader To allow loading firmware via serial port without
    the need For dedicated host PC software.
    
    On Reset, the loader waits 2 seconds For the serial port To enter the Bootloader.
    If so, the Flashloader runs. Otherwise, the Code performs a GoTo 4
    which must be a vector To the start of the user program.
    If the user Code space is blank, the Flashloader go To a Stop For waiting
    a new Reset from the user.

    The terminal must be configured For 57600, 38400, 19200, 9600 Bauds,
    NO Parity, 2 Stop Bits And uses Software flow control On/OFF.
    For Xtal >= 40Mhz, otherwize the terminal could be configured up To 115200 Bauds.

    This Bootloader does not need any modification of the Asm file executed by the MPLAB.

    The Code has not been tested with "ERASE_FLASH_BLOCKSIZE = 256" And "WRITE_FLASH_BLOCKSIZE = 128".

    The PIC Device ID number is tested before using the Bootloader. The PIC will
    be not flashed For a different PIC. A "DVF" error Code is sent To the terminal If
    the PIC is different than the Bootloader Code.

    This Bootloader will be flashed At the BOTTOM of the ROM memory.
    The lenght of the Bootloader is Variable according of the options choosen by the user.
    The Bootloader uses a doble UART buffer To increase communication speed And
    To facilitate a good decodification of the Hex lines.

    Baud Rate:
    Because the Bootloader communicates with the Terminal (PC side) with On/OFF protocol
    While the file is downloading, the Terminal software must To have the same baud rate As the Bootloader.

    Version 3.0
    In this new version the Bootloader Code is written in the lower part of the 
    ROM memory. Then the Boot block ($000000-$0007FFh) could be code-protected And write-protected.

    Version 4.0
    A Random acces Code is added To the Bootloader. (8 ASCII digits) The Code will
    change every time the download is done.
*) 
'===============================================================================
' REFERENCES:
' The idea of this Code is taken from the:
' PIC18 BootLoader version 0.63 with software flow control:
' 2013 by vegipete@strongedge.net
' 
' Some other ideas have been taking from the "umc_loader" included in the PDS.
' Author: David John Barker 04/08/2011
'
' Otherwize my code is completly different.
'===============================================================================
' This program is free software, you can redistribute it and/or modify.
' The code is offered as-is and the author take any responsibility for its use.
' This program is distributed in the hope that it will be useful, but WITHOUT ANY
' WARRANTY. However, a mention of the author and compiler would be appreciated.
'
' Este programa es un software libre, puede redistribuirlo y/o modificarlo.
' Se entrega este cdigo tal como esta y el autor NO se hace responsable de
' su utilizacin. Este programa se distribuye con la esperanza de que sea til,
' pero SIN NINGUNA GARANTA. No obstante citar el autor y el compilador
' sera muy apreciado.

' Ce programme est un logiciel libre, vous pouvez le redistribuer et/ou modifier.
' Ce logiciel est offert tel qu'il est, et son auteur NE prend AUCUNE responsabilit
' pour son utilization. Ce programme est distribu dans l'espoir qu'il sera util,
' mais SANS AUCUNE GARANTIE. Cependant mentionner l'auteur et le compileur serait
' trs aprci.
'-------------------------------------------------------------------------------
' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
' INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
' PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR OR COPYRIGHT
' HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
' OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
' SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'###############################################################################
'*******************************************************************************
' How to know the Device ID:
' Read the value from the "DeviceList.Inc" file, also from the PICxxxx.dev files
' in the Device MPLAB folder or from the datasheet.
'                  WRITE  ERASE
'    DEVICE        BLOCK  BLOCK
'    __________________________
'   | PIC18F65K40 |  64 |  64 |
'   | PIC18F66K40 |  64 |  64 |
'   | PIC18F67K40 | 128 | 128 |
'   |_____________|_____|_____|
'
' How to know the Device ID:
' Read the value from the "DeviceList.Inc" file, also from the PICxxxx.dev files
' in the Device MPLAB folder or from the datasheet.
'
' Read the ID from MPLAB folder:
' Transform the hex value (id=0x5540) to binary (%0101010101000000).
' Shift the value to the right 5 times (>>5).
' And transform the new value to decimal (%01010101010) => 682.
'*******************************************************************************
'* FOSC:                                                                       *
'*******************************************************************************
(*
    From the PDS Manual.
    18F device XTAL values:
    Declare Xtal 3, 4, 7, 8, 10, 12, 14, 16, 19, 20, 22, 24, 25, 29, 32, 33, 40, 48, or 64.
    Inform the compiler what frequency oscillator is being used.
    Some commands are very dependant on the oscillator frequency, Rsin, Rsout,
    DelayMs, and DelayUs being just a few. In order for the compiler to adjust
    the correct timing for these commands, it must know what frequency crystal
    is being used.
    The Xtal frequencies 3, 7, 14, 19 , 22 and 29 are for 3.58MHz, 7.2MHz, 14.32MHz, 19.66MHz,
    22.1184MHz and 29.2MHz respectively.

    BAUD RATE TABLE:
    Xtal avaibled: 64, 48, 40, 32, 25, 24, 20, 16, 14 (14.32MHz), 12, 10
    115200 Bauds FOSC = 64Mhz = tested
    115200 Bauds FOSC = 40Mhz = tested
    57600 Bauds FOSC = 40Mhz = tested   
    57600 Bauds FOSC = 20Mhz = tested
    57600 Bauds FOSC = 16Mhz = tested
    38400 Bauds FOSC = 10Mhz = tested
    19200 Bauds FOSC = 10Mhz = tested
    9600 Bauds FOSC = 10Mhz = tested   
*)
'-------------------------------------------------------------------------------   
'   BAUD RATE TABLE FOR TERMINAL:
'    __________________________________________   
'   | Declare Xtal      | Internal| Baud Rate  |
'   |                   |   OSC   | Bootloader |
'   |___________________|_________|____________|
'   | >= 40Mhz          |   NO    |  115200    |
'   | >= 40Mhz          |   YES   |   57600    |
'   | >= 16 And < 40Mhz |   -     |   57600    |
'   | >= 10 And < 16Mhz |   -     |   38400    |
'   |___________________|_________|____________|
'
'*******************************************************************************
' This software could be used for the following PICs with minimum modifications.
'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'###############################################################################
'*******************************************************************************
CONFIG01: ' (1) CONFIG BY USER:                                                *
'* DECLARE THE DEVICE                                                          *
'* DECLARE FOSC: Xtal: 64, 48, 40, 32, 25, 24, 20, 16, 14 (14.32MHz), 12,10    *
'* DECLARE THE PLL ON/OFF                                                      *
'* CHOOSE THE INTERNAL OSCILLATOR OR NOT                                       *
'* CHOOSE THE UART1, UART2, UART3 or UART4 FOR THE BOOTLOADER                  *
'******************************************************************************* 
'    '' The USER must copy this "Declare" in the user main project.
'    Device = 18F65K40
'    Declare Xtal = 64

'    '' DEFINE THE PLL: On / OFF
'    $define PLL_ConfigFuses On

'    '' USING THE INTERNAL OSCILLATOR:
'    '' Uncomment the next line to enable the configuration.
'    '' FOSC avaibled: 64, 48, 32, 16, 12 (The PLL_ConfigFuses is not necessary)
'    '$define _InternalOSC_

'    '' Enable the Line to use the CLOCK OUT FUNCTION: 
'    '' If CLKOUT function is enabled, Port function On RA6 is disabled.     
'    '$Define CLKOUT_Function

'    '' DEFINE the UART name used for the bootloader. (UART1, UART2, UART3, UART4)
'    $define UART_USED_FOR_BOOTLOADER UART1
'===============================================================================
'    '' The USER must copy this "Declare" in the user main project.
'    Device = 18F66K40
'    Declare Xtal = 64

'    '' DEFINE THE PLL: On / OFF
'    $define PLL_ConfigFuses On

'    '' USING THE INTERNAL OSCILLATOR:
'    '' Uncomment the next line to enable the configuration.
'    '' FOSC avaibled: 64, 48, 32, 16, 12 (The PLL_ConfigFuses is not necessary)
'    '$define _InternalOSC_

'    '' Enable the Line to use the CLOCK OUT FUNCTION: 
'    '' If CLKOUT function is enabled, Port function On RA6 is disabled.     
'    '$Define CLKOUT_Function

'    '' DEFINE the UART name used for the bootloader. (UART1, UART2, UART3, UART4)
'    $define UART_USED_FOR_BOOTLOADER UART1 
'===============================================================================
    '' The USER must copy this "Declare" in the user main project.
    Device = 18F67K40
    Declare Xtal = 64

    '' DEFINE THE PLL: On / OFF
    $define PLL_ConfigFuses On

    '' USING THE INTERNAL OSCILLATOR:
    '' Uncomment the next line to enable the configuration.
    '' FOSC avaibled: 64, 48, 32, 16, 12 (The PLL_ConfigFuses is not necessary)
    '$define _InternalOSC_

    '' Enable the Line to use the CLOCK OUT FUNCTION: 
    '' If CLKOUT function is enabled, Port function On RA6 is disabled.     
    '$Define CLKOUT_Function

    '' DEFINE the UART name used for the bootloader. (UART1, UART2, UART3, UART4 Or UART5)
    $define UART_USED_FOR_BOOTLOADER UART1 

'*******************************************************************************
CONFIG02: '* (2) CONFIG BY USER:                                               *
'* DECLARE SOME MORE OPTIONS FOR THE USER:                                     *
'*******************************************************************************
' Disable the "_EnablePrintDeviceID_" option if there is not enough place
' for the Boot sector to compile.
'
' (1) ENABLE THE MESSAGES OF ERROR OR INFO:
' Uncomment the next line to Enable the mensages for the terminal (E & W).
' Please choose this option as default.
    $define _EnableMessages_

' (2) ID LOCATIONS:
' Uncomment the next line to write the ID Locations option.
' Please choose this option as default.
    $define _EnableWriteIDLOCS_

' (3) ENABLE TO PRINT THE DEVICE ID TO THE TERMINAL:
' Uncomment the next line to print in the first line of the screen the Devide ID number and revision of the PIC. 
' Please choose this option as default.
    $define _EnablePrintDeviceID_
'    $define _EnablePrintRevisionID_

' (4) ENABLE TO USE THE RANDOM PASSWORD MODE 0: (Random passwords)
' Uncomment the next line to Enable a Random Password request before to download the .hex file.
'    $define _EnablePassWordMode0_

' (5) ENABLE TO PRINT 1000 RANDOM PASSWORDS:
' Uncomment the next line to allow to print 1000 RANDOM codes to the terminal. See label STEP10.
'    $define _EnableToPrintCodes_

' (6) ENABLE TO CHECK THE PASSWORD INTEGRITY IN EEPROM: (WITH MODE 0)
' Uncomment the next line to Check the integrity of the password in the eeprom memory (If erased). (optional)
'    $define _ControlPWEeprom_

' (7) ENABLE TO USE THE UNIQUE PASSWORD MODE 1: (Only one password)
' Uncomment the next line to Enable Only one Password defined at compilation using the KEY only.
'    $define _EnablePassWordMode1_

' (8) ENABLE THE PROTECTION OF THE BOOT BLOCK:
' Uncomment the next line to allow to protet the Boot Block against reading, writing, & Table read
    $define _EnableProtectBootBlock_

' (9) ERASE THE EEPROM MEMORY BEFORE TO WRITE AT DOWNLOADING TIME:
' Uncomment the next line to choose the option "ERASE THE EEPROM MEMORY BEFORE TO WRITE".
'    $define _EnableEraseEeprom_

' (10) ENABLE THE TERMINAL COMMAND TO ERASE THE EEPROM MEMORY ONLY: (Password + Command)
' Command: (:020000040050AA) After command, The number of password is not incremented.
' Uncomment the next line to erase the eeprom memory with a special command with the terminal. 
'    $define _EraseEepromCommand_
'
' (11) ENABLE TO ERASE the IDLOCS positions At compile time, And when this firmware
'      is loaded into the PIC by a programmer.
'    $define _EraseIDlocs_
'
' (12) Configure the StartTimeOut.
' ENABLE THE TIMEOUT FOR 5 Seconds, otherwize 2 Seconds.
'    $define StartTimeOut5Sec
'
'*******************************************************************************
CONFIG03: '* (3) CONFIG BY USER:                                               *
'* DEFINE THE RANGE TO ERASE THE EEPROM MEMORY (IF NEEDED)                     *
'*******************************************************************************
    ' Option "ERASE THE EEPROM MEMORY BEFORE TO WRITE". Configuration by user.
    ' Define the RANGE you want to erase the Eeprom Data.                  
    $define EEP_ERASE_LOW_ADDRESS 0 ' Declare the Low address of the eeprom memory to be erased. (MIN = 0)
    $define EEP_ERASE_HIGH_ADDRESS  _eeprom - 1   ' (max = _eeprom - 1) Declare the High address of the eeprom memory to be erased.
'*******************************************************************************
CONFIG04: '* (4)                                                               *
'* DECLARE THE KEYS FOR PASSWORDS                                              *
'* CONFIGURATION BY USER:                                                      *
'*******************************************************************************
    ' Definitions of the KEYS. (Different KEYs For every project)
    Symbol KEYNumber = $1DF4E178    ' Declare a KEY Number (8 HEX characters) to encode the Bootloader Password. (one different KEY for every project)
    Symbol KEYASCII = $6A           ' Declare another KEY Number (2 Hex characters) To encode the ASCII Bootloader Password written in the eeprom.
    Symbol PW_CONTROL = $55         ' Must be modified in every project. Declare a Value to control the eeprom integrity.
'###############################################################################
'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
'*******************************************************************************
CONFIG05: ' (5)                                                                *
'* INFORMATION FOR THE USER PROJECT:                                           *
'******************************************************************************* 
(*
 INFO FOR THE PDS USER:
 BOOTLOADER CODE:
 VERIFY THE ASM FILE:
 The Result must To be like this:

proton#code#start
    org 0
    goto proton#main#start
    nop
    nop
    org 8
    goto ISR_High
    org 24
    goto ISR_Low
proton#main#start
...
'-------------------------------------------------------------------------------
 USER Code: Insert 2 new lines in the Main User Code.
 Insert "Declare Bootloader = OFF" 
 Insert "Declare PROTON_START_ADDRESS = 2048" 
 Example:

In the basic code:
    Declare Xtal = 64
    Declare Optimiser_Level = 2
    Declare Dead_Code_Remove = On
    Declare Watchdog = On
    Declare Bootloader = Off
    Declare PROTON_START_ADDRESS = 2048
....

VERIFY THE ASM FILE:
The result must to be like this:

proton#code#start
    org 0X0800
    goto proton#main#start
    org 0X0808
    goto Interrupt_Handler_In   ' (if there is a High interrupt)
    org 0X0818
    goto ISR_Low                ' (If there is a Low Interrupt)
...
proton#main#start    
...
*)
'===============================================================================
' PLEASE, THE USER MUST READ THE DATASHEET TO WRITE THESE CONFIGURATIONS:   
'*******************************************************************************
' DECLARE THE CHARACTERISTICS OF THE PIC FOR THIS BOOTLOADER ONLY:
'*******************************************************************************
CONFIG06: '* (6) CONFIG BY USER:                                               *
'* DECLARE THE UART AND ALL PINS                                               *
'*******************************************************************************
'  From the datasheet find and write the correct values of the configuration parameters.
    '---------------------------------------------------------------------------
'    ' Declare the UART number used for the bootloader. 
'    $define UART_USED_FOR_BOOTLOADER UART1      ' Write UART1 = 0 OR UART2 = 1. Write the UART used for the bootloader. 
'===============================================================================
'===============================================================================
' Declare the UART1:
$if UART_USED_FOR_BOOTLOADER = UART1
    ' Write your parameters (read the datasheet & the .def / .PPI files)

'    $define UART_PORT PORTA                    ' Write the name of the PORT where the UART1 is connected.
'    $define UART_PORT PORTB                    ' Write the name of the PORT where the UART1 is connected.
    $define UART_PORT PORTC                    ' Write the name of the PORT where the UART1 is connected.
'    $define UART_PORT PORTD                    ' Write the name of the PORT where the UART1 is connected.
    '---------------------------------------------------------------------------
'    $define UART_ANSEL ANSEL                   ' Write the ANSEL SFR to set the UART1 PORT to digital. (if not available, then comment the line)
'    $define UART_ANSEL ANSELH                  ' Write the ANSEL SFR to set the UART1 PORT to digital. (if not available, then comment the line)
'    $define UART_ANSEL ANSELA                  ' Write the ANSEL SFR to set the UART1 PORT to digital. (if not available, then comment the line)
'    $define UART_ANSEL ANSELB                  ' Write the ANSEL SFR to set the UART1 PORT to digital. (if not available, then comment the line)
'    $define UART_ANSEL ANSELC                  ' Write the ANSEL SFR to set the UART1 PORT to digital. (if not available, then comment the line)
'    $define UART_ANSEL ANSELD                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not available then comment the line)
    '---------------------------------------------------------------------------
'    $define UART_ANSEL_TXbit ANSELCbits_ANSELC6  ' Define the ANSEL TX1 Bit. (from .def file of the compiler) (If not available, Then comment the Line)
'    $define UART_ANSEL_RXbit ANSELCbits_ANSELC7  ' Define the ANSEL RX1 Bit. (from .def file of the compiler) (If not available, Then comment the Line)
    '---------------------------------------------------------------------------
    $define UART_TX_TRISBIT TRISCbits_TRISC6   ' Define the Tris TX1 Bit. (from .def file of the compiler)      
    $define UART_RX_TRISBIT TRISCbits_TRISC7   ' Define the Tris RX1 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define OPEN_DRAINbit ODCONCBits_ODCC6     ' Define the the Open Drain pin for TX1 Bit. (from .def file of the compiler)
    $define SLEW_RATEbit SLRCONCBits_SLRC6     ' Define the the Slew Rate pin for TX1 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define _TXIF_ PIR3Bits_TX1IF               '  Write the TX interrupt flag bit for UART1. (from .def file of the compiler)
    $define _TXIE_ PIE3Bits_TX1IE               ' Write the TX interrupt enable bit for UART1. (from .def file of the compiler)
    $define _RCIF_ PIR3Bits_RC1IF               ' Write the RX interrupt flag Bit For UART1. (from .def file of the compiler)
    $define _RCIE_ PIE3Bits_RC1IE               ' Write the RX interrupt enable bit for UART1. (from .def file of the compiler)
    '###########################################################################
    ' Automatic settings: 
    $define _TXSTA_ TX1STA
    $define _RCSTA_ RC1STA
    $define _SPBRG_ SPBRG1
    $define _SPBRGH_ SPBRGH1
    $define _BAUDCON_ BAUDCON1
    $define _BRG16_ BAUDCON1Bits_BRG16
    $define _TXREG_ TX1REG
    $define _RCREG_ RC1REG
    $define _TRMT_ TX1STABits_TRMT
    $define _CREN_ RC1STABits_CREN
    $define _OERR_ RC1STABits_OERR
    $define _FERR_ RC1STABits_FERR
    '###########################################################################
$endif
'===============================================================================
' Declare the UART2:
' All is disabled if the UART2 is not avaible.
' For these PICs the UART2 is in PORTG.
$if _uart > 1
$if UART_USED_FOR_BOOTLOADER = UART2
    ' Write your parameters (read the datasheet)
'    $define UART_PORT PORTA                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTB                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTC                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTD                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTE                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTF                    ' Write the name of the PORT where the UART2 is connected.
    $define UART_PORT PORTG                    ' Write the name of the PORT where the UART2 is connected.
    '---------------------------------------------------------------------------
'    $define UART_ANSEL ANSEL                   ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELH                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELA                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELB                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELD                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELE                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELF                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
    $define UART_ANSEL ANSELG                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
    '---------------------------------------------------------------------------
    $define UART_ANSEL_TXbit ANSELGbits_ANSELG1  ' Define the ANSEL TX1 Bit. (from .def file of the compiler) (If not, Then comment the Line)
    $define UART_ANSEL_RXbit ANSELGbits_ANSELG2  ' Define the ANSEL RX1 Bit. (from .def file of the compiler) (If not, Then comment the Line)
    '---------------------------------------------------------------------------
    $define UART_TX_TRISBIT TRISGbits_TRISG1   ' Define the Tris TX2 Bit. (from .def file of the compiler) 
    $define UART_RX_TRISBIT TRISGbits_TRISG2   ' Define the Tris RX2 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define OPEN_DRAINbit ODCONGBits_ODCG1     ' Define the the Open Drain pin for TX1 Bit. (from .def file of the compiler)
    $define SLEW_RATEbit SLRCONGBits_SLRG1     ' Define the the Slew Rate pin for TX1 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define _TXIF_ PIR3Bits_TX2IF              ' Write the interrupt flag bit for UART2. (from .def file of the compiler)
    $define _TXIE_ PIE3Bits_TX2IE              ' Write the interrupt enable bit for UART2. (from .def file of the compiler)
    $define _RCIF_ PIR3Bits_RC2IF              ' Write the RX interrupt flag Bit For UART2. (from .def file of the compiler)
    $define _RCIE_ PIE3Bits_RC2IE              ' Write the RX interrupt enable bit for UART2. (from .def file of the compiler)
    '###########################################################################
    ' Automatic settings:
    $define _TXSTA_ TX2STA
    $define _RCSTA_ RC2STA      
    $define _SPBRG_ SPBRG2
    $define _SPBRGH_ SPBRGH2
    $define _BAUDCON_ BAUDCON2
    $define _BRG16_ BAUDCON2Bits_BRG16
    $define _TXREG_ TX2REG
    $define _RCREG_ RC2REG
    $define _TRMT_ TX2STABits_TRMT
    $define _CREN_ RC2STABits_CREN
    $define _OERR_ RC2STABits_OERR
    $define _FERR_ RC2STABits_FERR
    '###########################################################################
$endif             
$endif
'===============================================================================
' Declare the UART3:
' All is disabled if the UART3 is not avaible.
' For these PICs the UART3 is in PORTE.
$if _uart > 2
$if UART_USED_FOR_BOOTLOADER = UART3
    ' Write your parameters (read the datasheet)
'    $define UART_PORT PORTA                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTB                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTC                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTD                    ' Write the name of the PORT where the UART2 is connected.
    $define UART_PORT PORTE                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTF                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTG                    ' Write the name of the PORT where the UART2 is connected.
    '---------------------------------------------------------------------------
'    $define UART_ANSEL ANSEL                   ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELH                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELA                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELB                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELD                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
    $define UART_ANSEL ANSELE                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELF                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELG                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
    '---------------------------------------------------------------------------
    $define UART_ANSEL_TXbit ANSELEbits_ANSELE0  ' Define the ANSEL TX1 Bit. (from .def file of the compiler) (If not, Then comment the Line)
    $define UART_ANSEL_RXbit ANSELEbits_ANSELE1  ' Define the ANSEL RX1 Bit. (from .def file of the compiler) (If not, Then comment the Line)
    '---------------------------------------------------------------------------
    $define UART_TX_TRISBIT TRISEbits_TRISE0   ' Define the Tris TX2 Bit. (from .def file of the compiler) 
    $define UART_RX_TRISBIT TRISEbits_TRISE1   ' Define the Tris RX2 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define OPEN_DRAINbit ODCONEBits_ODCE0     ' Define the the Open Drain pin for TX1 Bit. (from .def file of the compiler)
    $define SLEW_RATEbit SLRCONEBits_SLRE0     ' Define the the Slew Rate pin for TX1 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define _TXIF_ PIR4Bits_TX3IF              ' Write the interrupt flag bit for UART2. (from .def file of the compiler)
    $define _TXIE_ PIE4Bits_TX3IE              ' Write the interrupt enable bit for UART2. (from .def file of the compiler)
    $define _RCIF_ PIR4Bits_RC3IF              ' Write the RX interrupt flag Bit For UART2. (from .def file of the compiler)
    $define _RCIE_ PIE4Bits_RC3IE              ' Write the RX interrupt enable bit for UART2. (from .def file of the compiler)
    '###########################################################################
    ' Automatic settings:
    $define _TXSTA_ TX3STA
    $define _RCSTA_ RC3STA      
    $define _SPBRG_ SPBRG3
    $define _SPBRGH_ SPBRGH3
    $define _BAUDCON_ BAUDCON3
    $define _BRG16_ BAUDCON3Bits_BRG16
    $define _TXREG_ TX3REG
    $define _RCREG_ RC3REG
    $define _TRMT_ TX3STABits_TRMT
    $define _CREN_ RC3STABits_CREN
    $define _OERR_ RC3STABits_OERR
    $define _FERR_ RC3STABits_FERR
    '###########################################################################
$endif          
$endif
'===============================================================================
' Declare the UART4:
' All is disabled if the UART2 is not avaible.
' For these PICs the UART2 is in PORTC.
$if _uart > 3
$if UART_USED_FOR_BOOTLOADER = UART4
    ' Write your parameters (read the datasheet)
'    $define UART_PORT PORTA                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTB                    ' Write the name of the PORT where the UART2 is connected.
    $define UART_PORT PORTC                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTD                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTE                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTF                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTG                    ' Write the name of the PORT where the UART2 is connected.
    '---------------------------------------------------------------------------
'    $define UART_ANSEL ANSEL                   ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELH                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELA                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELB                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELD                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELE                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELF                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELG                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
    '---------------------------------------------------------------------------
'    $define UART_ANSEL_TXbit ANSELCbits_ANSELC0  ' Define the ANSEL TX1 Bit. (from .def file of the compiler) (If not, Then comment the Line)
'    $define UART_ANSEL_RXbit ANSELCbits_ANSELC1  ' Define the ANSEL RX1 Bit. (from .def file of the compiler) (If not, Then comment the Line)
    '---------------------------------------------------------------------------
    $define UART_TX_TRISBIT TRISCbits_TRISC0   ' Define the Tris TX2 Bit. (from .def file of the compiler) 
    $define UART_RX_TRISBIT TRISCbits_TRISC1   ' Define the Tris RX2 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define OPEN_DRAINbit ODCONCBits_ODCC0     ' Define the the Open Drain pin for TX1 Bit. (from .def file of the compiler)
    $define SLEW_RATEbit SLRCONCBits_SLRC0     ' Define the the Slew Rate pin for TX1 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define _TXIF_ PIR4Bits_TX4IF              ' Write the interrupt flag bit for UART2. (from .def file of the compiler)
    $define _TXIE_ PIE4Bits_TX4IE              ' Write the interrupt enable bit for UART2. (from .def file of the compiler)
    $define _RCIF_ PIR4Bits_RC4IF              ' Write the RX interrupt flag Bit For UART2. (from .def file of the compiler)
    $define _RCIE_ PIE4Bits_RC4IE              ' Write the RX interrupt enable bit for UART2. (from .def file of the compiler)
    '###########################################################################
    ' Automatic settings:
    $define _TXSTA_ TX4STA
    $define _RCSTA_ RC4STA      
    $define _SPBRG_ SPBRG4
    $define _SPBRGH_ SPBRGH4
    $define _BAUDCON_ BAUDCON4
    $define _BRG16_ BAUDCON4Bits_BRG16
    $define _TXREG_ TX4REG
    $define _RCREG_ RC4REG
    $define _TRMT_ TX4STABits_TRMT
    $define _CREN_ RC4STABits_CREN
    $define _OERR_ RC4STABits_OERR
    $define _FERR_ RC4STABits_FERR
    '###########################################################################
$endif           
$endif
'===============================================================================
' Declare the UART5:
' All is disabled if the UART5 is not avaible.
' For these PICs the UART5 is in PORTE.
$if _uart > 4
$if UART_USED_FOR_BOOTLOADER = UART5
    ' Write your parameters (read the datasheet)
'    $define UART_PORT PORTA                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTB                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTC                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTD                    ' Write the name of the PORT where the UART2 is connected.
    $define UART_PORT PORTE                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTF                    ' Write the name of the PORT where the UART2 is connected.
'    $define UART_PORT PORTG                    ' Write the name of the PORT where the UART2 is connected.
    '---------------------------------------------------------------------------
'    $define UART_ANSEL ANSEL                   ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELH                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELA                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELB                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELD                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
    $define UART_ANSEL ANSELE                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELF                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
'    $define UART_ANSEL ANSELG                  ' Write the ANSEL SFR to set the UART2 PORT to digital. (if not exits then comment the line)
    '---------------------------------------------------------------------------
    $define UART_ANSEL_TXbit ANSELEbits_ANSELE2  ' Define the ANSEL TX1 Bit. (from .def file of the compiler) (If not, Then comment the Line)
    $define UART_ANSEL_RXbit ANSELEbits_ANSELE3  ' Define the ANSEL RX1 Bit. (from .def file of the compiler) (If not, Then comment the Line)
    '---------------------------------------------------------------------------
    $define UART_TX_TRISBIT TRISEbits_TRISE2   ' Define the Tris TX2 Bit. (from .def file of the compiler) 
    $define UART_RX_TRISBIT TRISEbits_TRISE3   ' Define the Tris RX2 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define OPEN_DRAINbit ODCONEBits_ODCE2     ' Define the the Open Drain pin for TX1 Bit. (from .def file of the compiler)
    $define SLEW_RATEbit SLRCONEBits_SLRE2     ' Define the the Slew Rate pin for TX1 Bit. (from .def file of the compiler)
    '---------------------------------------------------------------------------
    $define _TXIF_ PIR4Bits_TX5IF              ' Write the interrupt flag bit for UART2. (from .def file of the compiler)
    $define _TXIE_ PIE4Bits_TX5IE              ' Write the interrupt enable bit for UART2. (from .def file of the compiler)
    $define _RCIF_ PIR4Bits_RC5IF              ' Write the RX interrupt flag Bit For UART2. (from .def file of the compiler)
    $define _RCIE_ PIE4Bits_RC5IE              ' Write the RX interrupt enable bit for UART2. (from .def file of the compiler)
    '###########################################################################
    ' Automatic settings:
    $define _TXSTA_ TX5STA
    $define _RCSTA_ RC5STA      
    $define _SPBRG_ SPBRG5
    $define _SPBRGH_ SPBRGH5
    $define _BAUDCON_ BAUDCON5
    $define _BRG16_ BAUDCON5Bits_BRG16
    $define _TXREG_ TX5REG
    $define _RCREG_ RC5REG
    $define _TRMT_ TX5STABits_TRMT
    $define _CREN_ RC5STABits_CREN
    $define _OERR_ RC5STABits_OERR
    $define _FERR_ RC5STABits_FERR
    '###########################################################################
$endif            
$endif
'*******************************************************************************
CONFIG07: '* (7)                                                               *
'* DECLARE THE PIC FEATURES                                                    *
'* (Read the datasheet)                                                        *
'* CONFIGURATION BY USER For NEW PIC:                                          *
'*******************************************************************************
    '---------------------------------------------------------------------------
    ' Configure the StartTimeOut. Version 1.1.
    ' Enable this $Define to configure the Start Delay to 5 Seconds, otherwize 2 Seconds.
'    $define StartTimeOut5Sec
    '---------------------------------------------------------------------------
    ' Version 1.1.
    ' Declare the Address of the MyDummyByte variable in RAM after compiling.
    ' The values depend of the Erase Block Size and the Write Block Size of the PIC.
    ' For the UART_Buffer1 Address should be in BANK0_START.
    ' If :ERROR (IN ASM) THE UART BUFFER1 MUST RESIDE IN UPPER BANK) then decrease
    ' then "MyDummyByteAddress" value. 
    '
    ' CONFIGURATION BY USER FOR NEW PIC:
    $if _device = _18F65K40
        ' Read the Datasheet or MPLAB .dev files to know the Device ID value.
        $define __DEVICE_ID 27392
        ' Read the Datasheet to know the erase block value.
        $ifndef _erase
        $define _erase 64
        $endif
        ' Read the Datasheet to know the _bankAend value.
        $ifdef _bankAend
        $undef _bankAend
        $endif
        $define _bankAend 95 
        ' Read the PPI file to know the _uart value.
        $ifdef _uart
        $undef _uart
        $endif
        $define _uart 4
        ' Declare the Address of the MyDummyByte variable in RAM abd compile again.
        $define MyDummyByteAddress 76
    $endif

    $if _device = _18F66K40
        ' Read the Datasheet or MPLAB .dev files to know the Device ID value.
        $define __DEVICE_ID 27360
        ' Read the Datasheet to know the erase block value.
        $ifndef _erase
        $define _erase 64
        $endif
        ' Read the Datasheet to know the _bankAend value.
        $ifdef _bankAend
        $undef _bankAend
        $endif
        $define _bankAend 95 
        ' Read the PPI file to know the _uart value.
        $ifdef _uart
        $undef _uart
        $endif
        $define _uart 4
        ' Declare the Address of the MyDummyByte variable in RAM abd compile again.
        $define MyDummyByteAddress 76
    $endif

    $if _device = _18F67K40
        ' Read the Datasheet or MPLAB .dev files to know the Device ID value.
        $define __DEVICE_ID 27328
        ' Read the Datasheet to know the erase block value.
        $ifndef _erase
        $define _erase 128
        $endif
        ' Read the Datasheet to know the _bankAend value.
        $ifdef _bankAend
        $undef _bankAend
        $endif
        $define _bankAend 95 
        ' Read the PPI file to know the _uart value.
        $ifdef _uart
        $undef _uart
        $endif
        $define _uart 5
        ' Declare the Address of the MyDummyByte variable in RAM abd compile again.
        $define MyDummyByteAddress 76
    $endif
'###############################################################################
'*******************************************************************************
CONFIG08: '* (8) FOR DEVELOPER:                                                *
'* DECLARE THE CHARACTERISTICS OF THE PIC FOR THIS BOOTLOADER ONLY:            *
'* CONFIGURATIONS TAKEN FROM THE COMPILER. DO NOT MODIFY THESE LINES.          *
'*******************************************************************************
' DO NOT MODIFY THESE LINES: Configured by the DEVELOPER.
    ' The IDLOCS address is read from the asm file. Eight memory locations (200000h-200007h) are designated, see the datasheet
    ' The CONFIG address is read from the asm file. Location of CONFIG in memory map, see the datasheet  
    ' The ID address is read from the asm file. Location of (DEVID1) in memory map, see the datasheet
    '---------------------------------------------------------------------------
''    $define _eeprom_ 0                    ' ONLY FOR TESTING. (LINE RESERVED FOR DEVELOPER)
    $define _eeprom_ _eeprom                ' Read from the compiler.
    Symbol DEVICE_ID = __DEVICE_ID          ' Setting of the Device ID.
    $define WRITE_FLASH_BLOCKSIZE _block    ' Write Size of the Code Memory Write Segment (in Bytes), see the datasheet, read from the compiler.
    $define ERASE_FLASH_BLOCKSIZE _erase    ' Write Size of the Code Memory Erase Segment (in Bytes), (added in the .dev file in test)
    Symbol FLASH_MEMORY_SIZE = _code - 1    ' Size of the Code Memory Write Segment (in Bytes), read from the compiler.
    $define FLASH_64K_SIZE_TEST 65537       ' (65536 + 1) Constant to know the limit of flash Memory. (Fixed V1.1)
    Symbol EEP_ADDRESS = $310000            ' Declare the EEPROM address in Proton Hex format. Location of EEPROM in memory map, see the datasheet
    $define USER_PROTON_START_ADDRESS 2048  ' 2048 ($0800): (decimal values only) Declare Address of the beginning of the user code. (must to be the first byte of the erase block)
    $define _Xtal_MAX 64                    ' Declare the FOSC max of the PIC to control.
    '---------------------------------------------------------------------------
    ' Some Declares for the bootloader.
    Declare Optimiser_Level = 2
    Declare Dead_Code_Remove = On
    Declare Bootloader = OFF                 ' Declare Bootloader = OFF
    #ifndef watchdog_req
    #define watchdog_req 1                  ' Declare: Watchdog is always On for the bootloader.
    #endif                   
    $if FLASH_64K_SIZE_TEST < _code
    Declare Access_Upper_64K = On
    $endif
    '---------------------------------------------------------------------------
    ' Address In the second erase block of user memory to check the written code.  
    Symbol ADDRESS_USER_CODE_CHECK = USER_PROTON_START_ADDRESS + ERASE_FLASH_BLOCKSIZE + 10
    ' (2048 + 64 + 10 = 2122) Address In the 2nd erase block of user memory.
    '---------------------------------------------------------------------------
    ' Checking data of the .def file.
    Symbol CTL_WRITE_FLASH_BLOCKSIZE = WRITE_FLASH_BLOCKSIZE    ' Control the .def file, see ASM file.
    Symbol CTL_ERASE_FLASH_BLOCKSIZE = ERASE_FLASH_BLOCKSIZE    ' Control the .def file, see ASM file.
    Symbol CTL_CODE_FLASH_SIZE = _code                          ' Control the .def file, see ASM file.
    Symbol CTL_RAM_SIZE = _ram                                  ' Control the .def file, see ASM file.
    $ifdef _rambanks
    Symbol CTL_RAM_BANKS = _rambanks                            ' Control the .def file, see ASM file.
    $endif
    Symbol CTL_EEPROM_SIZE = _eeprom                            ' Control the .def file, see ASM file.
    Symbol CTL_BANKAEND_SIZE = _bankAend                        ' Control the .def file, see ASM file.
    Symbol CTL_NUMBER_UART = _uart                              ' Control the .def file, see ASM file.
    '---------------------------------------------------------------------------
    ' Automatic settings
    $ifdef _PMD4
    $define _Config0_Pins_SFR_    
    $endif
    $ifdef UART_ANSEL_TXbit
    $define _Config1_Pins_SFR_    
    $endif
    '---------------------------------------------------------------------------
    Symbol ORG_Proton_Start = USER_PROTON_START_ADDRESS   ' Proton Start Address for user code.
    Symbol ORG_ISR_High = (USER_PROTON_START_ADDRESS + 8) ' Proton Start Address for High Interrupt user code.
    Symbol ORG_ISR_Low = (USER_PROTON_START_ADDRESS + 24) ' Proton Start Address For Low Interrupt user Code. 
    '---------------------------------------------------------------------------
    ' Automatic settings (do not modify)
    ' Address in eeprom memory of the Password system. Conditional compilation for eeprom addresses.
    $define EEP_ADDR_PASSWORD_SYSTEM  _eeprom - 16  ' Define the Address in eeprom memory of the Password system. (better no change)
    '---------------------------------------------------------------------------
    ' Automatic settings
    $ifdef _EnablePassWordMode0_
        $if EEP_ERASE_HIGH_ADDRESS > EEP_ADDR_PASSWORD_SYSTEM - 1 ' Last Address available for the user of the eeprom memory.
            $undef EEP_ERASE_HIGH_ADDRESS
            $define EEP_ERASE_HIGH_ADDRESS EEP_ADDR_PASSWORD_SYSTEM - 1
            $define EEP_ERASE_ADDRESS_MAX EEP_ADDR_PASSWORD_SYSTEM - 1
        $else
            $define EEP_ERASE_ADDRESS_MAX EEP_ERASE_HIGH_ADDRESS
        $endif
    $else
        $if EEP_ERASE_HIGH_ADDRESS > _eeprom - 1    
            $undef EEP_ERASE_HIGH_ADDRESS    
            $define EEP_ERASE_HIGH_ADDRESS _eeprom - 1
            $define EEP_ERASE_ADDRESS_MAX _eeprom - 1   ' Last Address available for the user of the eeprom memory.
        $else
            $define EEP_ERASE_ADDRESS_MAX EEP_ERASE_HIGH_ADDRESS        
        $endif
    $endif
    '---------------------------------------------------------------------------
    Symbol CTL_EEP_ERASE_LOW_ADDR = EEP_ERASE_LOW_ADDRESS            ' Check the value of EEP Low Address in asm file.
    Symbol CTL_EEP_ERASE_HIGH_ADDR = EEP_ERASE_ADDRESS_MAX           ' Check the value of EEP High Address in asm file.
    Symbol EEP_ADDR_NB_PASSWORD = EEP_ADDR_PASSWORD_SYSTEM           ' Address in eeprom memory of the BCD counters (5xBCD).
    Symbol EEP_ADDR_PW_CONTROL = (EEP_ADDR_PASSWORD_SYSTEM +5)       ' Address in eeprom memory of the Password control (1xBIN).
    Symbol EEP_ADDR_PASSWORD_ENABLE = (EEP_ADDR_PASSWORD_SYSTEM + 6) ' Address in eeprom memory of the Enable Code (1xBIN).
    Symbol EEP_ADDR_PASSWORD = (EEP_ADDR_PASSWORD_SYSTEM + 7)        ' Address in eeprom memory of the Password (8xASCII).
    '---------------------------------------------------------------------------
    ' Calculate the value of the Dummy Array dimension.
    Symbol DummyArrayLenght = _bankAend - MyDummyByteAddress 
'###############################################################################
'*******************************************************************************
'* WRITE HERE THE USER CONGIG FUSES WITH THE "FUSE CONFIGURATOR" PLUGIN:       *
'*******************************************************************************
' WACHTDOG FOR PIC18FxxK22 SERIES:
'    WDTEN = ON     : "Always Enabled"
'    WDTEN = SWON   : "WDT is controlled by SWDTEN bit of the WDTCON register"
'    WDTEN = NOSLP  : "WDT is disabled in sleep, otherwise enabled. SWDTEN bit has no effect"
'    WDTEN = OFF    : "Really Disabled"
'
' WACHTDOG FOR PIC18Fxxxx SERIES:
'    WDTEN = ON     : "Always Enabled"
'    WDTEN = OFF    : "WDT is controlled by SWDTEN bit of the WDTCON register"
'*******************************************************************************
CONFIG09: '* (9) CONFIG BY USER:                                               *
'* WRITE THE CONFIG FUSES                                                      *
'*******************************************************************************
'    ' Enable the Line to use the CLOCK OUT FUNCTION: 
'    ' CLKOUT function is enabled, Port function On RA6 is disabled.     
'    '$Define CLKOUT_Function
'===============================================================================
' For Bootloader security with code.
' CPD = On	     ;Data EEPROM code-protected (version 4.0)
' CPB = On	     ;Boot block (000000-0007FFh) code-protected (version 4.0)
' WRTB = On	     ;Boot Block (000000-0007FFh) write-protected (version 4.0)
' EBRTB = On	 ;Boot Block (000000-0007FFh) protected from table reads executed in other blocks
' WRTC = On	     ;Configuration registers (300000-3000FFh) write-protected
'
$if _device = _18F65K40
Config_Start
$ifdef _InternalOSC_
    FEXTOSC = OFF	            ; External Oscillator not enabled
    RSTOSC = HFINTOSC_64MHZ	    ; HFINTOSC with HFFRQ = 64 MHz and CDIV = 1:1
    $ifdef CLKOUT_Function
    CLKOUTEN = On	        ;CLKOUT: FOSC/4, function is enabled, Port function on RA6 is disabled. 
    $else
    CLKOUTEN = OFF	        ;CLKOUT function is disabled, Port function on RA6 is enabled. 
    $endif    
$else
    $if PLL_ConfigFuses = On
        $if (_xtal >= 10) And (_xtal <= 64)
        FEXTOSC = HS	        ; HS (crystal oscillator) above 8 MHz; PFM set to high power
        RSTOSC = EXTOSC_4PLL	; EXTOSC with 4x PLL, with EXTOSC operating per FEXTOSC bits
        $else
            $error "Error in (Declare Xtal) with external Xtal"
        $endif
    $endif
    $if PLL_ConfigFuses = OFF
        $if (_xtal >= 10) And (_xtal <= 25)
        FEXTOSC = HS	    ; HS (crystal oscillator) above 8 MHz; PFM set to high power
        RSTOSC = EXTOSC     ; EXTOSC operating per FEXTOSC Bits (device manufacturing default)
        $else
            $if (_xtal > 25) And (_xtal <= 64)        
            FEXTOSC = ECH	    ; ECH  External Clock High-Power mode (above 8 MHz)
            RSTOSC = EXTOSC     ; EXTOSC operating per FEXTOSC Bits (device manufacturing default)        
            $else
                $error "Error in (Declare Xtal) with external Xtal"
            $endif
        $endif
        $ifdef CLKOUT_Function
        CLKOUTEN = On	        ;CLKOUT: FOSC/4, function is enabled, Port function on RA6 is disabled. 
        $else
        CLKOUTEN = OFF	        ;CLKOUT function is disabled, Port function on RA6 is enabled. 
        $endif  
    $endif
$endif
    CSWEN = On	     ;Writing to NOSC and NDIV is allowed
    FCMEN = OFF	    ;Fail-Safe Clock Monitor disabled
    MCLRE = EXTMCLR	;If LVP = 0, MCLR pin is MCLR; If LVP = 1, RE3 pin function is MCLR 
    PWRTE = On	     ;Power up timer enabled
    LPBOREN = On	;ULPBOR enabled
    BOREN = On	     ;Brown-out Reset enabled according to SBOREN
    BORV = VBOR_2P45	;Brown-out Reset Voltage (VBOR) set to 2.45V
    ZCD = OFF	    ;ZCD disabled. ZCD can be enabled by setting the ZCDSEN bit of ZCDCON
    PPS1WAY = OFF	;PPSLOCK bit can be set and cleared repeatedly (subject to the unlock sequence)
    STVREN = On	    ;Stack full/underflow will cause Reset
    Debug = OFF	    ;Background debugger disabled
    XINST = OFF	    ;Extended Instruction Set and Indexed Addressing Mode disabled
    WDTCPS = WDTCPS_5	;Divider ratio 1:1024
    WDTE = SWDTEN	    ;WDT enabled/disabled by SWDTEN bit
    WDTCWS = WDTCWS_7	;window always open (100%); software control; keyed access not required
    WDTCCS = SC	;Software Control
    WRT0 = OFF	;Block 0 (000800-001FFFh) not write-protected
    WRT1 = OFF	;Block 1 (002000-003FFFh) not write-protected
    WRT2 = OFF	;Block 2 (004000-005FFFh) not write-protected
    WRT3 = OFF	;Block 3 (006000-007FFFh) not write-protected
    WRTC = On	;Configuration registers (300000-30000Bh) write-protected
    $ifdef _EnableProtectBootBlock_
    WRTB = On	;Boot Block (000000-0007FFh) write-protected
    $else
    WRTB = OFF	;Boot Block (000000-0007FFh) not write-protected    
    $endif
    WRTD = OFF	;Data EEPROM not write-protected
    SCANE = On	;Scanner module is available for use, SCANMD bit can control the module
    LVP = OFF	;HV on MCLR/VPP must be used for programming.
    Cp = OFF	;UserNVM code protection disabled
    CPD = OFF	;DataNVM code protection disabled
    EBTR0 = OFF	;Block 0 (000800-001FFFh) not protected from table reads executed in other blocks
    EBTR1 = OFF	;Block 1 (002000-003FFFh) not protected from table reads executed in other blocks
    EBTR2 = OFF	;Block 2 (004000-005FFFh) not protected from table reads executed in other blocks
    EBTR3 = OFF	;Block 3 (006000-007FFFh) not protected from table reads executed in other blocks
    $ifdef _EnableProtectBootBlock_
    EBTRB = On	;Boot Block (000000-0007FFh) protected from table reads executed in other blocks
    $else
    EBTRB = OFF	;Boot Block (000000-0007FFh) not protected from table reads executed in other blocks
    $endif
Config_End
$endif
'-------------------------------------------------------------------------------
$if _device = _18F66K40
Config_Start
$ifdef _InternalOSC_
    FEXTOSC = OFF	            ; External Oscillator not enabled
    RSTOSC = HFINTOSC_64MHZ	    ; HFINTOSC with HFFRQ = 64 MHz and CDIV = 1:1
    $ifdef CLKOUT_Function
    CLKOUTEN = On	        ;CLKOUT: FOSC/4, function is enabled, Port function on RA6 is disabled. 
    $else
    CLKOUTEN = OFF	        ;CLKOUT function is disabled, Port function on RA6 is enabled. 
    $endif    
$else
    $if PLL_ConfigFuses = On
        $if (_xtal >= 10) And (_xtal <= 64)
        FEXTOSC = HS	        ; HS (crystal oscillator) above 8 MHz; PFM set to high power
        RSTOSC = EXTOSC_4PLL	; EXTOSC with 4x PLL, with EXTOSC operating per FEXTOSC bits
        $else
            $error "Error in (Declare Xtal) with external Xtal"
        $endif
    $endif
    $if PLL_ConfigFuses = OFF
        $if (_xtal >= 10) And (_xtal <= 25)
        FEXTOSC = HS	    ; HS (crystal oscillator) above 8 MHz; PFM set to high power
        RSTOSC = EXTOSC     ; EXTOSC operating per FEXTOSC Bits (device manufacturing default)
        $else
            $if (_xtal > 25) And (_xtal <= 64)        
            FEXTOSC = ECH	    ; ECH  External Clock High-Power mode (above 8 MHz)
            RSTOSC = EXTOSC     ; EXTOSC operating per FEXTOSC Bits (device manufacturing default)        
            $else
                $error "Error in (Declare Xtal) with external Xtal"
            $endif
        $endif
        $ifdef CLKOUT_Function
        CLKOUTEN = On	        ;CLKOUT: FOSC/4, function is enabled, Port function on RA6 is disabled. 
        $else
        CLKOUTEN = OFF	        ;CLKOUT function is disabled, Port function on RA6 is enabled. 
        $endif  
    $endif
$endif
    CSWEN = On	     ;Writing to NOSC and NDIV is allowed
    FCMEN = OFF	    ;Fail-Safe Clock Monitor disabled
    MCLRE = EXTMCLR	;If LVP = 0, MCLR pin is MCLR; If LVP = 1, RE3 pin function is MCLR 
    PWRTE = On	     ;Power up timer enabled
    LPBOREN = On	;ULPBOR enabled
    BOREN = On	     ;Brown-out Reset enabled according to SBOREN
    BORV = VBOR_2P45	;Brown-out Reset Voltage (VBOR) set to 2.45V
    ZCD = OFF	      ;ZCD disabled. ZCD can be enabled by setting the ZCDSEN bit of ZCDCON
    PPS1WAY = OFF	;PPSLOCK bit can be set and cleared repeatedly (subject to the unlock sequence)
    STVREN = On	        ;Stack full/underflow will cause Reset
    Debug = OFF	        ;Background debugger disabled
    XINST = OFF	        ;Extended Instruction Set and Indexed Addressing Mode disabled
    WDTCPS = WDTCPS_5	;Divider ratio 1:1024
    WDTE = SWDTEN	   ;WDT enabled/disabled by SWDTEN bit
    WDTCWS = WDTCWS_7	;window always open (100%); software control; keyed access not required
    WDTCCS = SC	        ;Software Control
    WRT0 = OFF	;Block 0 (000800-003FFFh) not write-protected
    WRT1 = OFF	;Block 1 (004000-007FFFh) not write-protected
    WRT2 = OFF	;Block 2 (008000-00BFFFh) not write-protected
    WRT3 = OFF	;Block 3 (00C000-00FFFFh) not write-protected
    WRTC = On	;Configuration registers (300000-30000Bh) write-protected
    $ifdef _EnableProtectBootBlock_
    WRTB = On	;Boot Block (000000-0007FFh) write-protected
    $else
    WRTB = OFF	;Boot Block (000000-0007FFh) not write-protected    
    $endif
    WRTD = OFF	;Data EEPROM not write-protected
    SCANE = On	;Scanner module is available for use, SCANMD bit can control the module
    LVP = OFF	;HV on MCLR/VPP must be used for programming.
    Cp = OFF	;UserNVM code protection disabled
    CPD = OFF	;DataNVM code protection disabled
    EBTR0 = OFF	;Block 0 (000800-003FFFh) not protected from table reads executed in other blocks
    EBTR1 = OFF	;Block 1 (004000-007FFFh) not protected from table reads executed in other blocks
    EBTR2 = OFF	;Block 2 (008000-00BFFFh) not protected from table reads executed in other blocks
    EBTR3 = OFF	;Block 3 (00C000-00FFFFh) not protected from table reads executed in other blocks
    $ifdef _EnableProtectBootBlock_
    EBTRB = On	;Boot Block (000000-0007FFh) protected from table reads executed in other blocks
    $else
    EBTRB = OFF	;Boot Block (000000-0007FFh) not protected from table reads executed in other blocks 
    $endif
Config_End
$endif
'-------------------------------------------------------------------------------
$if _device = _18F67K40
Config_Start
$ifdef _InternalOSC_
    FEXTOSC = OFF	            ; External Oscillator not enabled
    RSTOSC = HFINTOSC_64MHZ	    ; HFINTOSC with HFFRQ = 64 MHz and CDIV = 1:1
    $ifdef CLKOUT_Function
    CLKOUTEN = On	        ;CLKOUT: FOSC/4, function is enabled, Port function on RA6 is disabled. 
    $else
    CLKOUTEN = OFF	        ;CLKOUT function is disabled, Port function on RA6 is enabled. 
    $endif    
$else
    $if PLL_ConfigFuses = On
        $if (_xtal >= 10) And (_xtal <= 64)
        FEXTOSC = HS	        ; HS (crystal oscillator) above 8 MHz; PFM set to high power
        RSTOSC = EXTOSC_4PLL	; EXTOSC with 4x PLL, with EXTOSC operating per FEXTOSC bits
        $else
            $error "Error in (Declare Xtal) with external Xtal"
        $endif
    $endif
    $if PLL_ConfigFuses = OFF
        $if (_xtal >= 10) And (_xtal <= 25)
        FEXTOSC = HS	    ; HS (crystal oscillator) above 8 MHz; PFM set to high power
        RSTOSC = EXTOSC     ; EXTOSC operating per FEXTOSC Bits (device manufacturing default)
        $else
            $if (_xtal > 25) And (_xtal <= 64)        
            FEXTOSC = ECH	    ; ECH  External Clock High-Power mode (above 8 MHz)
            RSTOSC = EXTOSC     ; EXTOSC operating per FEXTOSC Bits (device manufacturing default)        
            $else
                $error "Error in (Declare Xtal) with external Xtal"
            $endif
        $endif
        $ifdef CLKOUT_Function
        CLKOUTEN = On	        ;CLKOUT: FOSC/4, function is enabled, Port function on RA6 is disabled. 
        $else
        CLKOUTEN = OFF	        ;CLKOUT function is disabled, Port function on RA6 is enabled. 
        $endif  
    $endif
$endif
    CSWEN = On	     ;Writing to NOSC and NDIV is allowed
    FCMEN = OFF	;Fail-Safe Clock Monitor disabled
    MCLRE = EXTMCLR	;If LVP = 0, MCLR pin is MCLR; If LVP = 1, RE3 pin function is MCLR 
    PWRTE = On	;Power up timer enabled
    LPBOREN = On	;ULPBOR enabled
    BOREN = On	;Brown-out Reset enabled according to SBOREN
    BORV = VBOR_2P45	;Brown-out Reset Voltage (VBOR) set to 2.45V
    ZCD = OFF	;ZCD disabled. ZCD can be enabled by setting the ZCDSEN bit of ZCDCON
    PPS1WAY = On	;PPSLOCK bit can be cleared and set only once; PPS registers remain locked after one clear/set cycle
    STVREN = On	;Stack full/underflow will cause Reset
    Debug = OFF	;Background debugger disabled
    XINST = OFF	;Extended Instruction Set and Indexed Addressing Mode disabled
    WDTCPS = WDTCPS_5	;Divider ratio 1:1024
    WDTE = SWDTEN	;WDT enabled/disabled by SWDTEN bit
    WDTCWS = WDTCWS_7	;window always open (100%); software control; keyed access not required
    WDTCCS = SC	;Software Control
    WRT0 = OFF	;Block 0 (000800-003FFFh) not write-protected
    WRT1 = OFF	;Block 1 (004000-007FFFh) not write-protected
    WRT2 = OFF	;Block 2 (008000-00BFFFh) not write-protected
    WRT3 = OFF	;Block 3 (00C000-00FFFFh) not write-protected
    WRT4 = OFF	;Block 4 (010000-013FFFh) not write-protected
    WRT5 = OFF	;Block 5 (014000-017FFFh) not write-protected
    WRT6 = OFF	;Block 6 (018000-01BFFFh) not write-protected
    WRT7 = OFF	;Block 7 (01C000-01FFFFh) not write-protected
    WRTC = On	;Configuration registers (300000-30000Bh) write-protected
    $ifdef _EnableProtectBootBlock_
    WRTB = On	;Boot Block (000000-0007FFh) write-protected
    $else
    WRTB = OFF	;Boot Block (000000-0007FFh) not write-protected    
    $endif
    WRTD = OFF	;Data EEPROM not write-protected
    SCANE = On	;Scanner module is available for use, SCANMD bit can control the module
    LVP = OFF	;HV on MCLR/VPP must be used for programming.
    Cp = OFF	;UserNVM code protection disabled
    CPD = OFF	;DataNVM code protection disabled
    EBTR0 = OFF	;Block 0 (000800-003FFFh) not protected from table reads executed in other blocks
    EBTR1 = OFF	;Block 1 (004000-007FFFh) not protected from table reads executed in other blocks
    EBTR2 = OFF	;Block 2 (008000-00BFFFh) not protected from table reads executed in other blocks
    EBTR3 = OFF	;Block 3 (00C000-00FFFFh) not protected from table reads executed in other blocks
    EBTR4 = OFF	;Block 4 (010000-013FFFh) not protected from table reads executed in other blocks
    EBTR5 = OFF	;Block 5 (014000-017FFFh) not protected from table reads executed in other blocks
    EBTR6 = OFF	;Block 6 (018000-01BFFFh) not protected from table reads executed in other blocks
    EBTR7 = OFF	;Block 7 (01C000-01FFFFh) not protected from table reads executed in other blocks
    $ifdef _EnableProtectBootBlock_
    EBTRB = On	;Boot Block (000000-0007FFh) protected from table reads executed in other blocks
    $else
    EBTRB = OFF	;Boot Block (000000-0007FFh) not protected from table reads executed in other blocks
    $endif
Config_End
$endif
'*******************************************************************************
'* IF WRITING IDLOCS IS NEEDED THEN COPY THIS TABLE IN YOUR MAIN PROGRAM IN    *
'* THE END OF YOUR CODE .                                                      *
'* WRITE YOUR VALUES...                                                        *
'*******************************************************************************
'*******************************************************************************
CONFIG10: '* (10) CONFIG BY USER:                                              *
'* WRITE THE ID LOCS OPTIONS                                                   *
'* Copy & paste this code in your main code and modify it.                     *
'*******************************************************************************
(*
(ASCII characters or NUMBERS)
'-------------------------------------------------------------------------------
Asm
__IDLOCS   0x200000,"L"     ; Or 0x01  (your numbers)
__IDLOCS   0x200001,"O"     ; Or 0x02
__IDLOCS   0x200002,"A"     ; Or 0x03
__IDLOCS   0x200003,"D"     ; Or 0x04
__IDLOCS   0x200004,"E"     ; Or 0x05
__IDLOCS   0x200005,"R"     ; Or 0x06
__IDLOCS   0x200006,"5"     ; Or 0x07
__IDLOCS   0x200007,"0"     ; Or 0x08
EndAsm
'-------------------------------------------------------------------------------
*)    
'###############################################################################
'*******************************************************************************
CONFIG11: '* (11)                                                              *
'* CHECKING THE CONFIGURATION FOR THE DEVELOPER: DO NOT MODIFY                 *
'*******************************************************************************
    ' Check if the device is a PIC18.
    $if _core <> 16
        $error "This Bootloader code is for the PIC18 device ONLY!"
    $endif
'-------------------------------------------------------------------------------
    ' Check if the flash is avaible to read & write.
    $if _flash <> 1
        $error "This Device has NOT the Self Modifying support!"
    $endif
'-------------------------------------------------------------------------------
    ' Check the FOSC.
    $if _xtal > _Xtal_MAX
        $error "NO SUCH FOSC in this device!"    
    $else    
        $if (_xtal <> 64) And (_xtal <> 48) And (_xtal <> 40) And (_xtal <> 32) And (_xtal <> 25) And (_xtal <> 24) And (_xtal <> 20) And (_xtal <> 16) And (_xtal <> 14) And (_xtal <> 12) And (_xtal <> 10)
            $error "NO such FOSC for this Bootloader!"
        $endif
    $endif   
    '---------------------------------------------------------------------------
    ' Version 2.1
    ' Check the Internal Oscillator.
    $ifdef _InternalOSC_
        $if _xtal > 64
            $error "NO such FOSC for the internal oscillator!"
        $endif
        $if _BaudRate > 57600
        $undef _BaudRate
        $endif
    $endif
'-------------------------------------------------------------------------------
    ' Configuration Delay (25MS) for oscillator startup.
    $if _xtal = 64
    $define _OSC_Delay 45000
    $endif
    $if _xtal = 48
    $define _OSC_Delay 34000
    $endif
    $if _xtal = 40
    $define _OSC_Delay 28000
    $endif    
    $if _xtal = 32
    $define _OSC_Delay 25500
    $endif
    $if _xtal = 25
    $define _OSC_Delay 17500
    $endif
    $if _xtal = 24
    $define _OSC_Delay 16900
    $endif
    $if _xtal = 20
    $define _OSC_Delay 14000
    $endif
    $if _xtal = 16
    $define _OSC_Delay 11200
    $endif
    $if _xtal = 14
    $define _OSC_Delay 10000
    $endif
    $if _xtal = 12
    $define _OSC_Delay 8430
    $endif    
    $if _xtal = 10
    $define _OSC_Delay 7000
    $endif

    ' Calculate the Oscillator Delay for Internal and External oscillator.
    $ifdef _InternalOSC_
    Symbol INTOscillatorDelay = _OSC_Delay / 3     ' Value for internal oscillator. 25mS (more or less) Could be modified by the user.
    $else
    Symbol INTOscillatorDelay = _OSC_Delay / 4 ' Value for external Xtal oscillator.    
    $endif
'-------------------------------------------------------------------------------
    $ifndef NVMCON1bits_NVMREG0
        $error "The Bootloader code is NOT compatible with this PIC!"
    $endif 
    $ifndef NVMCON1bits_NVMREG1
        $error "The Bootloader code is NOT compatible with this PIC!"
    $endif
'-------------------------------------------------------------------------------
' Xtal avaibled: 64, 48, 40, 32, 25, 24, 20, 16, 14 (14.32MHz), 12, 10
' 115200 Bauds FOSC = 64Mhz = tested
' 115200 Bauds FOSC = 40Mhz = tested
' 57600 Bauds FOSC = 40Mhz = tested   
' 57600 Bauds FOSC = 20Mhz = tested
' 57600 Bauds FOSC = 16Mhz = tested
' 57600 Bauds FOSC = 10Mhz = NOT runs
' 38400 Bauds FOSC = 10Mhz = tested
' 19200 Bauds FOSC = 10Mhz = tested
' 9600 Bauds FOSC = 10Mhz = tested     
'
' The Declare UART Baudrate is automatic depending of the Xtal.
    $ifndef _BaudRate
        $if _xtal >= 40
            $ifdef _InternalOSC_
            $define _BaudRate 57600
            $else
            $define _BaudRate 115200
            $endif
        $endif
        $if (_xtal >= 16) And (_xtal < 40)
            $define _BaudRate 57600
        $endif
        $if _xtal < 16
            $define _BaudRate 38400
        $endif
    $endif
'-------------------------------------------------------------------------------
    ' Check the UART features.
    $if _uart = 0
        $error "NO UART avaible in this Device!"
    $endif
    $if UART_USED_FOR_BOOTLOADER +1 > _uart
        $error "NO such UART in this Device!"
    $endif
    $if (_BaudRate <> 115200) And (_BaudRate <> 57600) And (_BaudRate <> 38400) And (_BaudRate <> 19200) And (_BaudRate <> 9600)
        $error "You must to declare one speed for the UART!"
    $endif

    $if !_defined(UART_PORT)
        $error "The UART PORT is not defined!"
    $endif
    $if !_defined(UART_TX_TRISBIT)
        $error "The UART TX TRIS bit is not defined!"   ' Modified V2.0
    $endif
    $if !_defined(UART_RX_TRISBIT)
        $error "The UART RX TRIS bit is not defined!"   ' Modified V2.0
    $endif
    $if !_defined(_TXIF_)
        $error "The TX UART interrupt flag bit (TXIF) is not defined!"
    $endif
    $if !_defined(_TXIE_)
        $error "The TX UART interrupt flag enabled (TXIE) is not defined!"
    $endif
    $if !_defined(_RCIF_)
        $error "The RX UART interrupt flag bit is not defined!"
    $endif
    $if !_defined(_RCIE_)
        $error "The RX UART interrupt flag enabled is not defined!"
    $endif
'-------------------------------------------------------------------------------
' Testing the Baudrate. Is it correct?
    $if _BaudRate = 115200
    $if (_xtal <> 64) And (_xtal <> 48) And (_xtal <> 40)
        $error "NO such FOSC for this Baud rate!"
    $endif    
    $endif

    $if _BaudRate = 57600
    $if (_xtal <> 64) And (_xtal <> 48) And (_xtal <> 40) And (_xtal <> 32) And (_xtal <> 25) And (_xtal <> 24) And (_xtal <> 20) And (_xtal <> 16)
        $error "NO such FOSC for this Baud rate!"
    $endif    
    $endif    

    $if _BaudRate = 38400
    $if (_xtal <> 64) And (_xtal <> 48) And (_xtal <> 40) And (_xtal <> 32) And (_xtal <> 25) And (_xtal <> 24) And (_xtal <> 20) And (_xtal <> 16) And (_xtal <> 14) And (_xtal <> 12) And (_xtal <> 10)
        $error "NO such FOSC for this Baud rate!"
    $endif    
    $endif
    
    $if _BaudRate = 19200
    $if (_xtal <> 64) And (_xtal <> 48) And (_xtal <> 40) And (_xtal <> 32) And (_xtal <> 25) And (_xtal <> 24) And (_xtal <> 20) And (_xtal <> 16) And (_xtal <> 14) And (_xtal <> 12) And (_xtal <> 10)
        $error "NO such FOSC for this Baud rate!"
    $endif     
    $endif
    
    $if _BaudRate = 9600
    $if (_xtal <> 64) And (_xtal <> 48) And (_xtal <> 40) And (_xtal <> 32) And (_xtal <> 25) And (_xtal <> 24) And (_xtal <> 20) And (_xtal <> 16) And (_xtal <> 14) And (_xtal <> 12) And (_xtal <> 10)
        $error "NO such FOSC for this Baud rate!"
    $endif     
    $endif            
'-------------------------------------------------------------------------------
    $ifndef __DEVICE_ID
        $error "The Device ID is not defined!"
    $endif
'-------------------------------------------------------------------------------
    $ifndef _erase
        $error "The ERASE BLOCK value is not defined!"
    $endif
'------------------------------------------------------------------------------- 
    ' Check the Write Block size.
    ' If your WRITE_FLASH_BLOCKSIZE is not correct, verify the PICxx.def and the PICxxx.PPI files in the PDS folder.
    $if WRITE_FLASH_BLOCKSIZE > ERASE_FLASH_BLOCKSIZE
        $error "WRITE_FLASH_BLOCKSIZE must be < or = to ERASE_FLASH_BLOCKSIZE! See the datasheet."
    $endif
'-------------------------------------------------------------------------------
    ' Check the Erase Block size. Verify your value from the datasheet.
    ' ERASE BLOCK values above 128 KB have not been tested. (could be correct)
    $if (ERASE_FLASH_BLOCKSIZE > 128)
        $error "This value of the ERASE_FLASH_BLOCKSIZE has not been tested! (Disable this message if needed)"  ' Modified V2.0
    $endif
    $if (ERASE_FLASH_BLOCKSIZE < 64)
        $error "The ERASE_FLASH_BLOCKSIZE value seems not correct! Check the datasheet"
    $endif
'-------------------------------------------------------------------------------
    ' Check the eeprom viability    
    $ifdef _EnablePassWordMode0_
    $define _ChangeAddEepMax_
    $endif

    $if _eeprom_ = 0
        $undef _EnablePassWordMode0_
        $undef _EnableEraseEeprom_
        $undef _EnableToPrintCodes_
        $undef _EraseEepromCommand_
        $undef _ControlPWEeprom_
    $else
        ' Check some possible errors of eeprom parameters.
        $if EEP_ERASE_LOW_ADDRESS = EEP_ERASE_HIGH_ADDRESS
            $error "ERROR! No such value for EEPROM address!"
        $endif
        $if EEP_ERASE_LOW_ADDRESS > EEP_ERASE_HIGH_ADDRESS
            $error "ERROR! The range of the LOW ADDRESS of data eeprom to be erased is not correct!"
        $endif
        $ifdef _ChangeAddEepMax_
            $if EEP_ERASE_LOW_ADDRESS > EEP_ERASE_ADDRESS_MAX - 1
                $error "ERROR! No such value for Low EEPROM address!"    
            $endif
            $if EEP_ERASE_HIGH_ADDRESS > EEP_ERASE_ADDRESS_MAX
                $error "ERROR! No such value for High EEPROM address!"    
            $endif
        $else
            $if EEP_ERASE_LOW_ADDRESS > _eeprom - 2
                $error "ERROR! No such value for Low EEPROM address!"    
            $endif
            $if EEP_ERASE_HIGH_ADDRESS > _eeprom - 1
                $error "ERROR! No such value for High EEPROM address!"    
            $endif        
        $endif
    $endif
'-------------------------------------------------------------------------------
    ' Check the password option 
    $if _defined(_EnablePassWordMode0_) And _defined(_EnablePassWordMode1_)
        $error "ERROR! 2 Password modes cannot be enabled"
    $endif

    $if USER_PROTON_START_ADDRESS <> 2048
        $error "ERROR! The 'USER_PROTON_START_ADDRESS' must be equal to 2048!"
    $endif
    
    ' Disable all options if _EnableToPrintCodes_ is enabled. (There is not enough space in the flash memory)
    $ifdef _EnableToPrintCodes_
        $ifndef _EnablePassWordMode0_
            $error "ERROR! The Password Mode0 must be enabled!"
        $endif
        $ifdef _EnableEraseEeprom_
            $undef _EnableEraseEeprom_
        $endif
        $ifdef _EraseEepromCommand_
            $undef _EraseEepromCommand_
        $endif 
        $ifdef _EnablePrintDeviceID_
            $undef _EnablePrintDeviceID_
        $endif
    $endif

    ' Enable/Disable the commands.
    $ifdef _EraseEepromCommand_
        $if !_defined(_EnablePassWordMode0_) And !_defined(_EnablePassWordMode1_)
            $error "ERROR! One Password Mode for the (_EraseEepromCommand_) must be enabled!"
        $endif
        $define _EnableCommand_
    $endif

    $if _defined(_EnablePassWordMode0_) Or _defined(_EnablePassWordMode1_)
        $ifndef _EnableMessages_
        $define _EnableMessages_
        $endif
    $endif
'-------------------------------------------------------------------------------
    $if PLL_ConfigFuses != On And PLL_ConfigFuses != OFF
        $error "ERROR! in the PLL_ConfigFuses Declare"
    $endif     
'-------------------------------------------------------------------------------
'    $if !_defined(_EnablePrintDeviceID_) And _defined(_EnablePrintRevisionID_)
'    $undef _EnablePrintRevisionID_
'    $endif
'*******************************************************************************
CONFIG12: '* (12)                                                              *
'* BOOTLOADER MACROS: RESERVED For DEVELOPER                                   *
'*******************************************************************************
$define READ2BYTES_CKS() GoSub Read2BYTES_CheckSum_Sub
    
$define EraseEEPROM() GoSub EraseEEPROM_Sub

$define EraseROM() GoSub EraseUserROM_Sub

$define Conv2HEX_1BINByte() GoSub TWOHEX_1BIN_Sub

$define EEPROMWrite() GoSub Make_EEPROMWrite_Sub

$define Read_DeviceID() GoSub ReadDeviceID_Sub

$if FLASH_64K_SIZE_TEST > _code
$define HRSOutSTRG_CRLF(pString) '
    TBLPTRW = pString             '
    GoSub HRSOUT#MSTR_CRLF_Sub

$define HRSOutSTRG(pString) '
    TBLPTRW = pString        '
    GoSub HRSOUT#MSTR_Sub

$else
$define HRSOutSTRG_CRLF(pString)  '
    Set_Bank TBLPTRL              '
    Asm                           '
    Movlw (pString & 255)         '
    Movwf TBLPTRL,0               '
    Movlw ((pString >> 8) & 255)  '
    Movwf TBLPTRH,0               '
    Movlw ((pString >> 16) & 255) '
    Movwf TBLPTRU,0               '
    EndAsm                        '
    GoSub HRSOUT#MSTR_CRLF_Sub

$define HRSOutSTRG(pString)       '
    Set_Bank TBLPTRL              '
    Asm                           '
    Movlw (pString & 255)         '
    Movwf TBLPTRL,0               '
    Movlw ((pString >> 8) & 255)  '
    Movwf TBLPTRH,0               '
    Movlw ((pString >> 16) & 255) '
    Movwf TBLPTRU,0               '
    EndAsm                        '
    GoSub HRSOUT#MSTR_Sub

$endif

$define XOFF()   '
    WREG = _XOFF '
    GoSub HRSOUT_Sub
    
$define XON()   '
    WREG = _XON '
    GoSub HRSOUT_Sub
              
$define Conv_LoNib_to_ASCII(pValue) '
    WREG = pValue                   '
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub

$define Conv1Byte_to_2ASCII_CRLF(pValue) '
    Byte_DSB = pValue                    '
    GoSub CONV_1BYTE_2HEX_Sub
    
$define Conv1Byte_to_2ASCII(pValue) '
    Byte_DSB = pValue               '
    GoSub CONV_1BYTE_2HEXBB_Sub

'*******************************************************************************
CONFIG13: '* (13)                                                              *
'* CREATE SOME PARAMETERS: RESERVED For DEVELOPER                              *
'*******************************************************************************
    $if ERASE_FLASH_BLOCKSIZE = 1024
    Symbol ERASE_BLOCK_SHIFT_BITS = 10
    Symbol MASK_ERASE_BLOCK = $FFFC00
    $endif

    $if ERASE_FLASH_BLOCKSIZE = 512
    Symbol ERASE_BLOCK_SHIFT_BITS = 9
    Symbol MASK_ERASE_BLOCK = $FFFE00
    $endif

    $if ERASE_FLASH_BLOCKSIZE = 256
    Symbol ERASE_BLOCK_SHIFT_BITS = 8
    Symbol MASK_ERASE_BLOCK = $FFFF00
    $endif

    $if ERASE_FLASH_BLOCKSIZE = 128
    Symbol ERASE_BLOCK_SHIFT_BITS = 7
    Symbol MASK_ERASE_BLOCK = $FFFF80
    $endif

    $if ERASE_FLASH_BLOCKSIZE = 64
    Symbol ERASE_BLOCK_SHIFT_BITS = 6
    Symbol MASK_ERASE_BLOCK = $FFFFC0
    $endif
'===============================================================================    
    $if WRITE_FLASH_BLOCKSIZE = 256
    Symbol WRITE_BLOCK_SHIFT_BITS = 8
    Symbol MASK_HEAD_ADDRESS_OFFSET = $0000FF
    Symbol MASK_WRITE_BLOCK = $FFFF00
    $endif

    $if WRITE_FLASH_BLOCKSIZE = 128
    Symbol WRITE_BLOCK_SHIFT_BITS = 7
    Symbol MASK_HEAD_ADDRESS_OFFSET = $00007F
    Symbol MASK_WRITE_BLOCK = $FFFF80
    $endif

    $if WRITE_FLASH_BLOCKSIZE = 64
    Symbol WRITE_BLOCK_SHIFT_BITS = 6
    Symbol MASK_HEAD_ADDRESS_OFFSET = $00003F
    Symbol MASK_WRITE_BLOCK = $FFFFC0
    $endif

    $if WRITE_FLASH_BLOCKSIZE = 32
    Symbol WRITE_BLOCK_SHIFT_BITS = 5
    Symbol MASK_HEAD_ADDRESS_OFFSET = $00001F
    Symbol MASK_WRITE_BLOCK = $FFFFE0
    $endif

    $if WRITE_FLASH_BLOCKSIZE = 16
    Symbol WRITE_BLOCK_SHIFT_BITS = 4
    Symbol MASK_HEAD_ADDRESS_OFFSET = $00000F
    Symbol MASK_WRITE_BLOCK = $FFFFF0
    $endif

    $if WRITE_FLASH_BLOCKSIZE = 8
    Symbol WRITE_BLOCK_SHIFT_BITS = 3
    Symbol MASK_HEAD_ADDRESS_OFFSET = $000007
    Symbol MASK_WRITE_BLOCK = $FFFFF8
    $endif
'*******************************************************************************
'* BLOCKS LIMITS TO ERASE/WRITE:                                               *
'******************************************************************************* 
    Symbol FIRST_BLOCK_TO_ERASE = USER_PROTON_START_ADDRESS >> ERASE_BLOCK_SHIFT_BITS
    Symbol LAST_BLOCK_TO_ERASE = FLASH_MEMORY_SIZE >> ERASE_BLOCK_SHIFT_BITS    
    Symbol FIRST_BLOCK_TO_WRITE = USER_PROTON_START_ADDRESS >> WRITE_BLOCK_SHIFT_BITS
    Symbol LAST_BLOCK_TO_WRITE = FLASH_MEMORY_SIZE >> WRITE_BLOCK_SHIFT_BITS
'*******************************************************************************
'* SEPARATE EVERY BYTE OF THE KEY TO WRITE TO EEPROM MEMORY:                   *
'******************************************************************************* 
    Symbol KEYNumberByte3 = ((KEYNumber >> 24) & 255)
    Symbol KEYNumberByte2 = ((KEYNumber >> 16) & 255)   
    Symbol KEYNumberByte1 = ((KEYNumber >> 8)  & 255)
    Symbol KEYNumberByte0 = (KEYNumber & 255)
'*******************************************************************************
'* DEFINE SOME CONSTANTS FOR THE TERMINAL COMMUNICATION:                       *
'*******************************************************************************    
    Symbol CR = 13      ' Decimal 13, Carriage Return
    Symbol LF = 10      ' Decimal 10, Line Feed
    Symbol ETX = 3      ' Decimal  3, End of Text
    Symbol _XON = 17    ' Decimal 17, Resume transmission DC1, Software Protocol XON
    Symbol _XOFF = 19   ' Decimal 19, Pause transmission DC3, Software Protocol XOFF
    Symbol FF = 12      ' Decimal 12, Form Feed. 
    Symbol NULL = 0     ' Decimal 0,  Null character. 
    Symbol ESC = 27     ' Decimal 27, Escape. 
'*******************************************************************************
CONFIG15: '* (15) DEVELOPER:                                                   *
'* DEFINE THE VARIABLES OF THE BOOTLOADER:                                     *
'*******************************************************************************
    ' Define the EEPROM address for eeprom.
    $if _eeprom_ > 256
    Dim EEP_EEADR As NVMADRL.Word
    $else
    Dim EEP_EEADR As NVMADRL   
    $endif
    Dim LOOP_EEADR As Word System 

    ' Some aliases
    Dim TBLPTRW As TBLPTRL.Word     
'    Dim TBLPTRW As TBLPTR.Word
    Dim FSR0 As FSR0L.Word
    Dim FSR1 As FSR1L.Word
    Dim FSR2 As FSR2L.Word

    Dim LOOPByte As Byte System
    Dim BOOTLOADER_FLAGS As Word System
    Dim HEX32_ChkSum As Byte System
    Dim NumberHEXBytes As Byte System
    Dim NbDataBytesAvailabled As Byte System
    Dim NextBlockToWrite As Word System
    Dim BlockToErase As Word System
    Dim RecordType As Byte System
    Dim CheckSumFrame As Byte System
    Dim DataByte As Byte System
    Dim DataByte2 As Byte System
    Dim DataWord As Word System
'    Dim DataDword As Dword System   
    Dim PICRevision As Word System
    Dim PICPartNumber As Word System
    Dim Byte_DSB As Byte System 
    Dim EnabledToRead As Byte System
    Dim FullBlock As Byte System
    Dim BytesCounter As Byte System
    Dim PostScaler As Byte System
'    Dim PassWordError As Byte System
    Dim NumberBCDCounters As Byte System 
        
    'Declare some variables for the UART Buffer
    Dim UARTBuff0_Head As Byte System           ' Pointer for the buffer0.
    Dim UARTBuff0_Tail As Byte System           ' Pointer for the buffer0.
    Dim UARTBuff1_Head As Byte System           ' Pointer for the buffer1.
    Dim UARTBuff1_Tail As Byte System           ' Pointer for the buffer1.
    Dim UARTBuff1_Tail_COP As Byte System       ' Pointer for the buffer1.

    Dim DataBin As Byte System
    Dim UART_Byte As Byte System
    Dim IDLOC0_Low As Byte System
    Dim IDLOC0_High As Byte System
    Dim IDLOC0_Upper As Byte System
    Dim Config1H_Low As Byte System
    Dim Config1H_High As Byte System
    Dim Config1H_Upper As Byte System
'-------------------------------------------------------------------------------
    ' Declare some special variables for the ROM pointer.
    ' I detected an issue with the compiler with the "Declare Bootloader = On" (Compiler V3.5.8.6.).
    ' The aliases with "At" do not run correctly.
    ' The workaround is to use some array as reference. 
    Dim MyArray1[5] As Byte
    Dim HEX32_ADDRESS As Dword At MyArray1#0        ' There is the ROM address read from the UART.
    Dim TBLPTR_HEX32 As Word At MyArray1#0          ' TBLPTRW (Word) is the Address value from the HEX file.

    Dim TBLPTRU_HEX32 As HEX32_ADDRESS.Byte2        ' TBLPTRU (Byte) is the Address value from the Hex file.
    Dim TBLPTR_HEX32_PREVIOUS As Word

    Dim MyArray2[5] As Byte
    Dim DwordNewBlockToWrite As Dword At MyArray2#0 ' To make some shift with 3 bytes.
    Dim ReadNewBlockToWrite As Word At MyArray2#0
    
    Dim ShiftRegArray[4] As Byte
    Dim SHIFT_REG3 As ShiftRegArray#3
    Dim SHIFT_REG2 As ShiftRegArray#2
    Dim SHIFT_REG1 As ShiftRegArray#1
    Dim SHIFT_REG0 As ShiftRegArray#0

    Dim TEMP_REG3A As Byte
    Dim TEMP_REG3B As Byte
    Dim TEMP_REG2A As Byte
    Dim TEMP_REG2B As Byte
    Dim NEW_REG0 As Byte
    
    Dim NumberOfPasswordArray[5] As Byte   
    Dim PasswordMode1Array[9] As Byte 
'------------------------------------------------------------------------------- 
' ALiases for bits.
    Dim F_EnableReadCKS     As BOOTLOADER_FLAGS.0  ' Disable to read the Checksum
    Dim F_EnableCheckCode   As BOOTLOADER_FLAGS.1  ' Ask for code before to download the file.
    Dim F_CarryFlag         As BOOTLOADER_FLAGS.2
    Dim F_StartLine         As BOOTLOADER_FLAGS.3  ' Enable to read the first character of the line (:). 
    Dim F_BlocksErased      As BOOTLOADER_FLAGS.4
    Dim F_StopDecodingLines As BOOTLOADER_FLAGS.5
    Dim F_BootLoader_Error  As BOOTLOADER_FLAGS.6
    Dim F_WritingROM        As BOOTLOADER_FLAGS.7
    Dim F_WritingFlash      As BOOTLOADER_FLAGS.8    
'------------------------------------------------------------------------------- 
' Others aliases for Timer0.
    ' (another timer could be configured)
    Dim F_TimerIE As PIE0Bits_TMR0IE
    Dim F_TimerIF As PIR0Bits_TMR0IF
    Dim F_TimerON As T0CON0bits_T0EN
    $define _TimerH_ TMR0H
    $define _TimerL_ TMR0L
    $define _TXCON_ T0CON0 
'*******************************************************************************
'* DEFINE THE POSITION OF THE ARRAYS IN THE RAM:                               *
'*******************************************************************************   
'*******************************************************************************
CONFIG16: '* (16)                                                              *
'* DEFINE THE POSITION OF THE ARRAYS IN THE RAM MEMORY                         *
'*******************************************************************************
' DO NOT WRITE VARIABLES BELOW THIS LINE:
'===============================================================================
    Dim MyDummyByte As Byte     ' This byte is to calculate the position of the Arrays
'===============================================================================
' This part of code has been modified V1.1
    $ifndef _bankAend
    $error "The _bankAend parameter is missing in the PIC configuration CONFIG07"
    $endif
'===============================================================================
(*
    It is more important that the Arrays are not in BankA_End.
    Choose the RAM Bank Size that must be enough for UARTBuffer0 lenght (46Bytes) + UARTBuffer1 lenght.
    (95) (for the moment) is the Address in RAM of BANKA_END, etc...
    (MyDummyByte) is the address of the last DIM Variable (see asm file) before the DummyArray.
*)
'###############################################################################  
    $if WRITE_FLASH_BLOCKSIZE < 16
        Symbol UART_Buffer1_Lenght = 20     ' (16 + 4)
    $else
        Symbol UART_Buffer1_Lenght = WRITE_FLASH_BLOCKSIZE + 4
    $endif
    Symbol UART_Buffer0_Lenght = 46         ' For one line for the UART Buffer0.
'===============================================================================
    ' Modification for Loader version Version 1.1
    Dim DummyArray[DummyArrayLenght] As Byte
    Dim UART_Buffer1[UART_Buffer1_Lenght] As Byte
    Dim UART_Buffer0[UART_Buffer0_Lenght] As Byte
'###############################################################################  
    ' The variables must reside within Bank0.
    ' The UART_Buffer1 must to be in upper Bank.
    #ifdef BankA_End
        #if (UART_Buffer1#0 <= BankA_End)
            #error "The UART_Buffer1 must reside in Bank0, check the MyDummyByteAddress parameter" 
        #endif      
    #endif
'===============================================================================
'*******************************************************************************
(*
    Block type 04 - set high address
    If the high byte is not zero, an impossible location has been requested - Range Error
    If the low byte = 0x20, an ID location has been requested - supported.
    If the low byte = 0x30, a CONFIG location has been requested - not supported: Range Error
    If the low byte = 0xF0, an EEPROM location has been requested - supported.
    For a device with > 64K of flash, any high address other than 0x00F0.
    (or 0x0000) is a Range Error

    Block type 00 - normal line of data
    If the running data address matches the TBLPTRW, add the next byte to the write buffer.
    Otherwise, add 0xFF to the write buffer in the head of tail of the line until the TBLPTRW catches up.
    THEN THE HEX32 FILE DATA MUST BE IN ASCENDING ORDER. 
    The user could write some ORG commands in the code file. Better these ORG
    could be written at the head of a write block.
    For that the user could use this command: Org (Value & -_CWRITE_BLOCK), Value = 1000 is the ROM address.

    Fill write buffer with WRITE_FLASH_BLOCKSIZE bytes of data.
    When the write buffer is full, write the bytes to flash.
    Current block may have exactly enough bytes, or too many or too few.
    
    ID Locations
    Eight memory locations ($200000-$200007) are designated as ID locations,
    where the user can store checksum or other code identification numbers.
    These locations are both readable and writable during normal execution through
    the TBLRD and TBLWT instructions or during program/verify.
    The ID locations can be read when the device is code-protected. 
*)
'###############################################################################
'#### BOOTLOADER GENERAL CODE ##################################################
'###############################################################################
'
' IT IS NOT RECOMMENDED YOU TO CHANGE THIS CODE UNLESS YOU KNOW WHAT YOU ARE DOING. 
'
'###############################################################################
' Declare the Interrupts jumps.

    On_Hardware_Interrupt GoTo ISR_High
    
    On_Low_Interrupt GoTo ISR_Low

'-------------------------------------------------------------------------------
' PIC Loader start: 
PIC#Loader#Boot:    ' DO NOT MODIFY THIS LABEL.
    #ifdef WatchDog_Req
    Clrwdt
    #endif
    TBLPTRU = 0         ' (Modified Version 1.1)
'*******************************************************************************
'* Name    : Init the system                                                   *
'* Purpose : Clear some SFR's to init the Bootloader                           *
'*******************************************************************************
    WDTCON0Bits_SWDTEN = 0      ' Disabled the Watchdog bit option.
'    WDTCON0Bits_SEN_WDTCON0 = 0 ' Disabled the Watchdog bit option.
    INTCON = 0                  ' Interrupts NOT used.
    WDTCON1 = %00000111         ' Window = 100%, WATCHDOG TIMER CONTROL REGISTER 1
'*******************************************************************************
CONFIG17: ' (17) CONFIG BY USER:                                               *
'* Configure the Internal Oscillator by the PDS USER.                          *
'* The result of FOSC must be in the range:                                    *
'* FOSC avaibled: 64, 48, 40, 32, 25, 24, 20, 16, 14 (14.32MHz), 12, 10        *
'*******************************************************************************
'* Name    : Set the OSCCON/OSCCON2/OSCTUNE SFRs                               *
'* Purpose : Configuration of the INTERNAL OSCILLATOR.                         *
'* These configuration is set for the FOSC at maximum frequency                *
'*******************************************************************************
' Running with Internal oscillator ONLY.
' CONFIG FUSES:
'    FEXTOSC = OFF	            ;Oscillator not enabled
'    RSTOSC = HFINTOSC_64MHZ    ;HFINTOSC with HFFRQ = 64 MHz and CDIV = 1:1
'    CLKOUTEN = OFF	            ;CLKOUT function is disabled
'    CSWEN = On	                ;Writing to NOSC and NDIV is allowed
'
$ifdef _InternalOSC_
INT_OSC:
' Write the number of bytes if some code is written.
' See the Datasheet to configure the internal oscillator.
    OSCCON1 = %01100000                 ' HFINTOSC is used for 64Mhz as start up.

    ' OSCFRQ: HFINTOSC FREQUENCY SELECTION REGISTER
    $if _xtal = 64
    OSCFRQ = %00001000
    $else
        $if _xtal = 48
        OSCFRQ = %00000111    
        $else    
            $if _xtal = 32
            OSCFRQ = %00000110    
            $else   
                $if _xtal = 16
                OSCFRQ = %00000101    
                $else        
                    $if _xtal = 12
                    OSCFRQ = %00000100    
                    $else  
                        $error "Error in (Declare Xtal) with Internal Oscillator"
                    $endif
                $endif                    
            $endif
        $endif                        
    $endif
    While OSCSTATBits_HFOR = 0 : Wend   ' HFINTOSC Oscillator Ready bit. 1 = The oscillator is ready to be used
$endif
'*******************************************************************************
'* Name    : DELAY for Internal/External Oscillator (Version 2.1)              *
'* Purpose : Delay for Oscillator to stabilize.                                *
'*         : Good for Internal and External oscillator.                        *
'*         : Delay 25mS+ max                                                   *
'*******************************************************************************
' After Reset, every time the OSCCON & OSCTUNE are written, they need a delay to stabilize.  
' Time for HF-INTOSC to Stabilize: 1mS, Time for PLL to Stabilize: 2mS
' Delay max for code: 25mS (more or less). INTOscillatorDelay Could be modified by the user.
'
OSC_DELAY:
    DataWord = 0
    Repeat
        #ifdef WatchDog_Req
        Clrwdt
        #endif
        Inc DataWord
    Until DataWord = INTOscillatorDelay
'*******************************************************************************
'* Name    : Config UART pins by PPS                                           *
'* Purpose : Set the PPS pins                                                  *
'*******************************************************************************
' Config the UART pins to digital.
    '---------------------------------------------------------------------------
$if UART_USED_FOR_BOOTLOADER = UART1
    ' Config the UART1 pins by PPS, default pins.
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 0   ' Unlock PPS
    RX1PPS = 23     ' RC7 => RX1 
    RC6PPS = 9      ' RC6 <= TX1
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 1   ' Lock PPS
$endif
$if UART_USED_FOR_BOOTLOADER = UART2
    ' Config the UART2 pins by PPS, default pins.
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 0   ' Unlock PPS
    RX2PPS = 50     ' RG2 => RX2
    RG1PPS = 22     ' RG1 <= TX2
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 1   ' Lock PPS
$endif
$if UART_USED_FOR_BOOTLOADER = UART3
    ' Config the UART3 pins by PPS, default pins.
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 0   ' Unlock PPS
    RX3PPS = 33     ' RE1 => RX3
    RE0PPS = 16     ' RE0 <= TX3
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 1   ' Lock PPS
$endif
$if UART_USED_FOR_BOOTLOADER = UART4
    ' Config the UART4 pins by PPS, default pins.
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 0   ' Unlock PPS
    RX4PPS = 17     ' RC1 => RX4
    RC0PPS = 18     ' RC0 <= TX4
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 1   ' Lock PPS
$endif
$if UART_USED_FOR_BOOTLOADER = UART5
    ' Config the UART5 pins by PPS, default pins.
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 0   ' Unlock PPS
    RX5PPS = 23     ' RE3 => RX5
    RE2PPS = 20     ' RE2 <= TX5
    PPSLOCK = $55
    PPSLOCK = $AA
    PPSLOCK.0 = 1   ' Lock PPS
$endif
'-------------------------------------------------------------------------------
    OPEN_DRAINbit = 0    ' NO Open Drain for TX1 pin.
    SLEW_RATEbit = 0     ' Normal Slew Rate for TX1 pin.
'===============================================================================    
' DO NOT modify: Automatic settings.
'*******************************************************************************
STEP16: '* (16) DEVELOPER:                                                     *
'* Name    : Setting the UART parameters                                       *
'*******************************************************************************  
SETTING_UARTX:
    _TXIE_ = 0      ' Disable TX interrupt for the UARTs.
'-------------------------------------------------------------------------------
$ifdef _Config0_Pins_SFR_
    ' UARTx block enabled. (only confirm to enable the UART blocks)
    $if UART_USED_FOR_BOOTLOADER = UART1
    PMD5.2 = 0      ' 0 = UART1 Module is enabled, Clock Source is connected, module draws digital power.
    $endif
    $if UART_USED_FOR_BOOTLOADER = UART2
    PMD5.3 = 0      ' 0 = UART2 Module is enabled, Clock Source is connected, module draws digital power.
    $endif
    $if UART_USED_FOR_BOOTLOADER = UART3
    PMD5.4 = 0      ' 0 = UART3 Module is enabled, Clock Source is connected, module draws digital power.
    $endif
    $if UART_USED_FOR_BOOTLOADER = UART4
    PMD5.5 = 0      ' 0 = UART4 Module is enabled, Clock Source is connected, module draws digital power.
    $endif
    $if UART_USED_FOR_BOOTLOADER = UART5
    PMD5.6 = 0      ' 0 = UART5 Module is enabled, Clock Source is connected, module draws digital power.
    $endif
$endif
$ifdef _Config1_Pins_SFR_      
    ' Put the UARTx bits to input and analog but bit by bit only.
    UART_ANSEL_RXbit = 0    ' RX bit to digital
    UART_ANSEL_TXbit = 0    ' TX bit to digital
$endif

    UART_RX_TRISBIT = 1     ' RX bit in input...
    UART_TX_TRISBIT = 0     ' TX bit in output...
    $ifdef _BAUDCON_
    _BAUDCON_ = 0
    $endif
'*******************************************************************************
'* Name    : Setting the BAUD RATE of the UART                                 *
'* Purpose : Configure the _BaudRate For the 2 UARTs                           *
'* Xtal avaibled: 64, 48, 40, 32, 25, 24, 20, 16, 14 (14.32MHz), 12, 10        *
'*******************************************************************************
' It's not recommended you change these options unless you know what you are doing...       
'
$if _BaudRate = 9600
'#if (Xtal == 64)       
'    $ifdef _BRG16_
'    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator
'    $endif
'    $ifdef _SPBRGH_
'    _SPBRGH_ = 0
'    $endif
'    _SPBRG_ = 103
'    _TXSTA_ = 32
'    _RCSTA_ = 144
'#endif
#if (Xtal == 64)       
    $ifdef _BRG16_
    _BRG16_ = 1         ' Enable the 16 bit Baud Rate Generator
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 6
    $endif
    _SPBRG_ = 130
    _TXSTA_ = 36        ' Enable transmit, BRGH = 1
    _RCSTA_ = 144
#endif
#if (Xtal == 48)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 77
    _TXSTA_ = 32        ' Enable transmit, BRGH = 0
    _RCSTA_ = 144
#endif
#if (Xtal == 40)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 64
    _TXSTA_ = 32
    _RCSTA_ = 144
#endif
#if (Xtal == 32)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 207
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 25)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 162
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 24)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 155
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 20)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 129
    _TXSTA_ = 36        ' TXEN = 1, BRGH = 1
    _RCSTA_ = 144
#endif

'#if (Xtal == 16)
'    $ifdef _BRG16_
'    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
'    $endif
'    $ifdef _SPBRGH_
'    _SPBRGH_ = 0
'    $endif
'    _SPBRG_ = 103
'    _TXSTA_ = 36        ' TXEN = 1, BRGH = 1
'    _RCSTA_ = 144
'#endif
#if (Xtal == 16)
    $ifdef _BRG16_
    _BRG16_ = 1         ' Enable the 16 bit Baud Rate Generator
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 1
    $endif
    _SPBRG_ = 160
    _TXSTA_ = 36        ' TXEN = 1, BRGH = 1
    _RCSTA_ = 144
#endif
#if (Xtal == 14)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 92
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 12)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 77
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 10)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 64
    _TXSTA_ = 36
    _RCSTA_ = 144
    #endif     
$endif
'-------------------------------------------------------------------------------
$if _BaudRate = 19200
#if (Xtal == 64)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 207
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 48)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 155
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif   
#if (Xtal == 40)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 129
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif 
#if (Xtal == 32)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 103
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif 
#if (Xtal == 25)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 80
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif 
#if (Xtal == 24)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 77
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif 
#if (Xtal == 20)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 64
    _TXSTA_ = 36        ' TXEN = 1, BRGH = 1
    _RCSTA_ = 144
#endif

'    #if (Xtal == 16)       
'    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
'    _SPBRG_ = 16
'    _SPBRGH_ = 0
'    _TXSTA_ = 36
'    _RCSTA_ = 144
'    #endif
#if (Xtal == 16)
    $ifdef _BRG16_
    _BRG16_ = 1         ' Enable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 207
    _TXSTA_ = 36        ' TXEN = 1, BRGH = 1
    _RCSTA_ = 144
#endif
#if (Xtal == 14)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 46
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 12)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 38
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 10)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif r 
    _SPBRG_ = 32
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
$endif
'-------------------------------------------------------------------------------
$if _BaudRate = 38400
#if (Xtal == 64)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 103
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 48)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 77
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 40)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 64
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 32)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 51
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 25)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 40
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 24)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 38
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif    
#if (Xtal == 20)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 32
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif  
#if (Xtal == 16)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 25
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 14)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 22
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 12)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 19
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 10)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 15
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif    
$endif
'-------------------------------------------------------------------------------
$if _BaudRate = 57600
#if (Xtal == 64)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 68
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 48)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 51
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 40)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 42
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 32)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 34
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 25)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 26
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 24)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 25
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 20)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 21
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif

'    #if (Xtal == 16)
'    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
'    _SPBRG_ = 16
'    _SPBRGH_ = 0
'    _TXSTA_ = 36
'    _RCSTA_ = 144
'    #endif
#if (Xtal == 16)
    $ifdef _BRG16_
    _BRG16_ = 1         ' Enable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 68
    _TXSTA_ = 36            ' TXEN = 1, BRGH = 1
    _RCSTA_ = 144
#endif
#if (Xtal == 14)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 15
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 12)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 12
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 10)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 10
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif    
$endif
'-------------------------------------------------------------------------------
$if _BaudRate = 115200
#if (Xtal == 64)       
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 34
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 48)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 25
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 40)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 21
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 32)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 16
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 25)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 13
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 24)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif   
    _SPBRG_ = 12
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 20)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif  
    _SPBRG_ = 10
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif

'    #if (Xtal == 16)
'    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
'    _SPBRG_ = 8
'    _SPBRGH_ = 0
'    _TXSTA_ = 36
'    _RCSTA_ = 144
'    #endif
#if (Xtal == 16)
    $ifdef _BRG16_
    _BRG16_ = 1         ' Enable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif
    _SPBRG_ = 34
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 14)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 7
    _TXSTA_ = 36
    _RCSTA_ = 144
#endif
#if (Xtal == 12)
    $ifdef _BRG16_
    _BRG16_ = 0         ' Disable the 16 bit Baud Rate Generator 
    $endif
    $ifdef _SPBRGH_
    _SPBRGH_ = 0
    $endif 
    _SPBRG_ = 6
    _TXSTA_ = 36
    _RCSTA_ = 144
    #endif
$endif
'*******************************************************************************
'* Name    : UART STARTS                                                       *
'* Purpose : Starts the UART communication                                     *
'*         : Sends a XOFF() command to disable the terminal.                   *
'*******************************************************************************
UART_STARTS:
    GoSub CleanUARTErrors_Sub   ' UART cleaning
    XOFF()  ' Disable the terminal. There is some stuff to do before receiving bytes.
'*******************************************************************************
'* Name    : TIMER0 SETTINGS                                                   *
'* Purpose : Configure the Timer0 (2 seconds) for the UART time out            *
'*         : The TIMER0 interrupt is not used.                                 *
'*******************************************************************************
' A PostScaler is need for the Timer0 to allows 2 seconds delay with FOSC = 64Mhz.
' The time out delay is not precise. It is not very important.
' The Timer is not reloaded to save some bytes of code. (it is a more simple code)
' Xtal avaibled: 64, 48, 40, 32, 25, 24, 20, 16, 14 (14.32MHz), 12, 10
'-------------------------------------------------------------------------------
' T0CON1: TIMER0 CONTROL REGISTER 1
'    bit 3-0 T0CKPS<3:0>: Prescaler Rate Select bit
'        1111 = 1:32768
'        1110 = 1:16384
'        1101 = 1:8192
'        1100 = 1:4096
'        1011 = 1:2048
'        1010 = 1:1024
'        1001 = 1:512
'        1000 = 1:256
'        0111 = 1:128
'        0110 = 1:64
'        0101 = 1:32
'        0100 = 1:16
'        0011 = 1:8
'        0010 = 1:4
'        0001 = 1:2
'        0000 = 1:1
'-------------------------------------------------------------------------------
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:256 prescaler value: 1,048Sec, x PostScaler 2 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 64
    $define PreScalerValue 8
    $define PostScalerPW 10
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 3    
    $else
        $define PostScalerValue 2    
    $endif
    $endif
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:128 prescale value: 0,699Sec, x PostScaler 3 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 48
    $define PreScalerValue 7
    $define PostScalerPW 15
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 7    
    $else
        $define PostScalerValue 3    
    $endif
    $endif    
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:64 prescale value: 0,319Sec, x PostScaler 5 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 40
    $define PreScalerValue 6
    $define PostScalerPW 25
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 16    
    $else
        $define PostScalerValue 5    
    $endif
    $endif  
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:256 prescale value: 2,097Sec, x PostScaler 1
    ' PostScalerPW for Password = 2x5 => 10 seconds.c
    $if _xtal = 32
    $define PreScalerValue 8
    $define PostScalerPW 5
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 3    
    $else
        $define PostScalerValue 1    
    $endif
    $endif  
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:64 prescale value: 0,671Sec, x PostScaler 3 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 25
    $define PreScalerValue 6
    $define PostScalerPW 15
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 8    
    $else
        $define PostScalerValue 3    
    $endif
    $endif 
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:64 prescale value: 0,699Sec, x PostScaler 3 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 24
    $define PreScalerValue 6
    $define PostScalerPW 15
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 7    
    $else
        $define PostScalerValue 3    
    $endif
    $endif 
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:32 prescale value: 0,419Sec, x PostScaler 5 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 20
    $define PreScalerValue 5
    $define PostScalerPW 25
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 12    
    $else
        $define PostScalerValue 5    
    $endif
    $endif 
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:128 prescale value: 2,097Sec, x PostScaler 1 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 16
    $define PreScalerValue 7
    $define PostScalerPW 5
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 3    
    $else
        $define PostScalerValue 1    
    $endif
    $endif 
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:16 prescale value: 0,292Sec, x PostScaler 7 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 14
    $define PreScalerValue 4
    $define PostScalerPW 35
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 17    
    $else
        $define PostScalerValue 7    
    $endif
    $endif     
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:32 prescale value: 0,699Sec, x PostScaler 3 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 12
    $define PreScalerValue 5
    $define PostScalerPW 15
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 7    
    $else
        $define PostScalerValue 3    
    $endif
    $endif
    ' Timer0 is configured As a 16-Bit timer/Counter, 1:16 prescale value: 0,419Sec, x PostScaler 5 
    ' PostScalerPW for Password = 2x5 => 10 seconds.
    $if _xtal = 10
    $define PreScalerValue 4
    $define PostScalerPW 25
    $ifdef StartTimeOut5Sec
        $define PostScalerValue 12    
    $else
        $define PostScalerValue 5    
    $endif
    $endif
    '---------------------------------------------------------------------------    
    ' Timer0 configuration:
    $ifdef PreScalerValue
    T0CON0 = %00010000                      ' 16 Bits Timer, Postscaler = 1:1, Timer0 = OFF
    T0CON1 = %01000000 | PreScalerValue     ' Timer is configured with clock FOSC/4.   
    F_TimerIE = 0                           ' Disable the Timer0 interrupts. The Timer interrupt flag is used with polling only.
    $else
        $error " A Timer0 Configuration is needed!"
    $endif
'*******************************************************************************
'* Name    : START BOOTLOADER                                                  *
'* Purpose : Initialize some variables. Disable the Terminal                   *
'*******************************************************************************
START_BOOTLOADER:
    BOOTLOADER_FLAGS = 0        ' Clear all flags used in the program, Bootloader Error included.
'*******************************************************************************
'* Name    : ASK for PASSWORD                                                  *
'* Purpose : Send a "PW?" character to the terminal                            *
'*******************************************************************************
    $ifndef _EnableToPrintCodes_
    $if _defined(_EnablePassWordMode0_) Or _defined(_EnablePassWordMode1_)
    WREG = "P"
    GoSub HRSOUT_Sub
    WREG = "W"
    GoSub HRSOUT_Sub

    $ifdef _EnablePassWordMode0_
    ' Print the number (in BCD) of the Password counter before downloading.
    EEP_EEADR = EEP_ADDR_NB_PASSWORD + 4    ' Read from the eeprom.
    LOOPByte = 5
    Repeat
        NVMCON1 = 0                          ' To read the eeprom only
        NVMCON1Bits_RD = 1
        WREG = NVMDAT
        GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub
        GoSub HRSOUT_Sub
        Dec LOOPByte
        Dec EEP_EEADR
    Until LOOPByte = 0
    $endif

    WREG = "?"
    GoSub HRSOUT_Sub
    GoSub HRSOUT#CR_Sub    
    $endif
    $endif
'*******************************************************************************
'* Name    : Read the ID Locations Address                                     *
'* Purpose : Load in RAM the address of the ID LOCATIONS from asm file.        *
'*           Less configuration for the user.                                  *
'*******************************************************************************
$ifdef _EnableWriteIDLOCS_
    Set_Bank IDLOC0_Low
    Asm
    Movlw Low (IDLOC0)
    Movwf IDLOC0_Low
    Movlw High (IDLOC0)
    Movwf IDLOC0_High
    Movlw upper (IDLOC0)
    Movwf IDLOC0_Upper
    EndAsm
$endif
'*******************************************************************************
'* Name    : Read the Config Fuses Locations Address                           *
'* Purpose : Load in RAM the address of the Config Fuses from asm file.        *
'*           Less configuration for the user.                                  *
'*******************************************************************************
    Set_Bank Config1H_Upper
    Asm
    Movlw upper (Config1H)
    Movwf Config1H_Upper
    EndAsm
'*******************************************************************************
    $ifdef _EnablePassWordMode1_
    GoSub LoadPassword1_Sub         ' Load the static KEY in an array of 8 bytes.
    $endif
    GoSub ClearUARTBuffer1_Sub      ' Fill the UART buffer1 with NOP's ($FF)
    EnabledToRead = 0               ' Disable to read the data.
'*******************************************************************************
'* Name    : READ THE HEX ASCII FRAME  (43 ASCII characters max per line)      *
'* Purpose : Read the ":" Character until the 2 Checksum bytes                 *
'*         : Each line of the HEX file is followed by $0D, $0A.                *
'*         : Wait Until a (LF or CR) character is received.                    *
'*******************************************************************************
STARTLINE:
    PostScaler = PostScalerValue
STARTLINE33:
'-------------------------------------------------------------------------------
    UARTBuff0_Head = 0      ' Clear the pointer for the RX ASCII buffer.
    F_StartLine = 0         ' This flag is set when the ":" character is detected.
' Start the Timer parameters for the Time Out. (one line)
    _TimerH_ = 0            ' Timer High Byte value to start.
    Nop                     ' Some PICs K series need a Nop to load the 2 counters of the Timer.
    _TimerL_ = 0            ' Timer Low Byte value To start.
    Nop
    F_TimerON = 1           ' TIMER starts.
    F_TimerIF = 0           ' Clear the Timer interrupt flag. (in case of...)
'-------------------------------------------------------------------------------
    XON()                   ' Enable downloading the HEX file.
'=============================================================================== 
STARTLINE55:
    Repeat                  ' START reading one line.
        '=======================================================================
        ' Time Out delay for the UART input and receiving the bytes. The Time Out is set for every line to receive.
        '=======================================================================        
        While 1 = 1
            '-------------------------------------------------------------------
            If _RCIF_ = 1 Then          ' Receiving some byte from the UART?
                UART_Byte = _RCREG_     ' Copy the Received Byte the more faster as possible.
                Break                   ' Then copy the byte to the buffer.
            EndIf
            '-------------------------------------------------------------------
            ' UART Time Out
            If F_TimerIF = 1 Then                       ' Check the Timer Interrupt flag (polling)
                F_TimerIF = 0
                Dec PostScaler                          ' Decrement the PostScaler after every Timer interrupt.
                If PostScaler = 0 Then
                    If F_StopDecodingLines = 0 Then     ' Check if it has been a fatal error before that.
                        GoTo TIMEOUT                    ' It is a correct Time Out. 
                    Else
                        ' Exit by an error. Because any block could be written with bad code then...
                        GoTo WAITING_RESET_TIMEOUT_ERR  ' Version 3.1, goto erase the 2nd Block of ROM
                    EndIf
                EndIf            
            EndIf
            '-------------------------------------------------------------------
            #ifdef watchdog_req
            Clrwdt
            #endif
            '-------------------------------------------------------------------
        Wend
        '=======================================================================
        ' Check an error of the UART reception
        '======================================================================= 
        If _OERR_ = 1 Then UART_ERROR       ' Check for overrun error 
        If _FERR_ = 1 Then UART_ERROR       ' Check from framing errors
        '=======================================================================
        ' Detect the starting character of the HEX line
        '=======================================================================
        ' Avoid to load some trash before the first character.
        If F_StartLine = 0 Then             ' Enabled until receive ":".
            If UART_Byte <> ":" Then        ' Looking for the line start
                GoTo STARTLINE55            ' Discard different characters of ":". Waiting bytes again...
            EndIf
            F_StartLine = 1
        EndIf
        '=======================================================================
        ' Step 1 - Start line ":" (1 ASCII Character) load the buffer & exit
        ' After the final line detected (CR) exit faster to send the XOFF.
        '======================================================================= 
        If UART_Byte = CR Then ENDLINE              ' Exit directly, more faster.
'        If UART_Byte = LF Then ENDLINE              ' Exit.
        UART_Buffer0[UARTBuff0_Head] = UART_Byte    ' Loading every byte into the UART_Buffer0.
        Inc UARTBuff0_Head                          ' Increment the Buffer pointer for loading the next byte.
        '======================================================================= 
    Until UARTBuff0_Head = UART_Buffer0_Lenght      ' Check the limit for security
'===============================================================================
' UART ERROR: NO chance. The reception does not stop.
'=============================================================================== 
UART_ERROR:
    ' ERROR: UART error or overryde the limit of bytes OR the code did not read the CR/LF/UART error.
    GoSub CleanUARTErrors_Sub           ' UART cleaning
    If F_StopDecodingLines = 0 Then     ' Check if it has been a fatal error before that.
        GoTo SYSTEMERROR
    Else
        GoTo STARTLINE                  ' Reading all bytes and do nothing until the file closes.
    EndIf
'===============================================================================
' Disable downloading the HEX file for a while.
'===============================================================================
ENDLINE:
    XOFF()                              ' Disable the terminal. Now there is time To manage the all bytes of the received Line.
''    UART_Buffer0[UARTBuff0_Head] = CR   ' Not used in this version. Now there is more time to load a final character (CR) in UART_Buffer0.
    F_TimerON = 0                       ' Stop the Timer after reading a line.
    If F_StopDecodingLines = 1 Then
        GoTo STARTLINE                  ' One error has been already checked then wait the end of the file.
    EndIf
'*******************************************************************************
'* Name    : DECODE THE PASSWORD                                               *
'*******************************************************************************
$if _defined(_EnablePassWordMode0_) Or _defined(_EnablePassWordMode1_) 
CHECK_PASSWORD:
    If F_EnableCheckCode = 0 Then
        If UARTBuff0_Head = 9 Then           ' The file must have 9 characters included the ":", Carriage Return not included.
            #ifdef watchdog_req
            Clrwdt
            #endif
            '-------------------------------------------------------------------
            ' Check the incoming Password Mode 0
            $ifdef _EnablePassWordMode0_
            '-------------------------------------------------------------------
            $ifdef _ControlPWEeprom_
            ' Check the eeprom integrity.
            EEP_EEADR = EEP_ADDR_PW_CONTROL ' Address of the last byte of the eeprom
            NVMCON1 = 0                      ' Read only.
            NVMCON1Bits_RD = 1               ' Eeprom read.
            If NVMDAT <> PW_CONTROL Then
                GoTo SYSTEMERROR            ' Some problem with the Password control: Error!
            EndIf            
            $else
            NVMCON1 = 0                      ' Read only for next line of code.
            $endif
            '-------------------------------------------------------------------
            ' Check if it is the first time to read the password?
            EEP_EEADR = EEP_ADDR_PASSWORD_ENABLE
            NVMCON1Bits_RD = 1              ' Eeprom read.
            PRODL = NVMDAT                  ' If NVMDAT = $FF => first time then Password = :AAAAAAAA
            '===================================================================
            ' Read and compare the passwords.
            EEP_EEADR = EEP_ADDR_PASSWORD   ' Address of the upper byte of password.
            LOOPByte = 1                    ' Do not read the ":" character.
            Repeat                          ' Read and compare the Password.
                PRODH = UART_Buffer0[LOOPByte]      ' Read the incomimg password.
                NVMCON1 = 0                         ' Repeat read only because the NVMCON1 SFR could be corrupted.
                NVMCON1Bits_RD = 1                  ' Eeprom read.
                If PRODL <> $FF Then                ' Is it the first time to read the password?
                    NVMDAT = NVMDAT ^ KEYASCII      ' No, It does not, then restore the correct value.
                EndIf
                If NVMDAT <> PRODH Then             ' Compare the Password in the eeprom with the incomimg password.
                    GoTo TIMEOUT00                  ' Goto Check if there is the old user code?  Error!
                EndIf
                Inc LOOPByte
                Inc EEP_EEADR                       ' Read the next byte of eeprom password.
            Until LOOPByte > 8       
            '-------------------------------------------------------------------
            $else
            '-------------------------------------------------------------------
            ' Now check the incoming Password Mode 1. 
            LOOPByte = 0                            ' Read the ":" character.
            Repeat                                  ' Read and compare the Password.
                DataByte = UART_Buffer0[LOOPByte]           ' Read the incomimg password.
                DataByte2 = PasswordMode1Array[LOOPByte]    ' Read the local password.
                If DataByte <> DataByte2 Then               ' Compare the Password in ROM with the incomimg password.
                    GoTo TIMEOUT00                          ' Goto Check if there is the old user code?
                EndIf
                Inc LOOPByte
            Until LOOPByte > 8
            '-------------------------------------------------------------------
            $endif 
            '===================================================================
            ' The Password is correct now download the file.
            PostScaler = PostScalerPW   ' Load a new TIMEOUT for the file x 5 (= 10 seconds). Change it if necessary.
            F_EnableCheckCode = 1       ' Do not run this code again until the hex file is downloaded.
            GoTo STARTLINE33            ' Goto download the hex file.
        Else
            GoTo TIMEOUT00  ' The lenght of the Password is not correct, then it is an error.
        EndIf
    EndIf
$endif
'*******************************************************************************
CONFIG18: '* (18) READ BY USER:                                                *
'* Name    : PRINT LIST OF 1000 PASSWORDS OF THIS BOOTLOADER.                  *
'*******************************************************************************
' (1) Enable the followings lines ($Define _EnableToPrintCodes_) and (_EnablePassWordMode0_)
'     in STEP04 to print the passwords of this bootloader.
' (2) Compile the code.
' (3) Load this firmware into the PIC with an ICSP programmer.
' (4) Reset the PIC.
' (5) Entry the code (:AAAAAAAA), "SEND" the string twice.
' (6) 1000 different passwords are printed to your terminal.
' (7) Save all codes in a file.
' (8) Disable theses lines in normal operation ('$define _EnableToPrintCodes_) in STEP04.
' (9) Compile again for normal operation.
' (10) Load this firmware into the PIC with an ICSP programmer again.
'
$ifdef _EnableToPrintCodes_
    ' Print the Project Number or KEY.
    GoSub HRSOUT#CR_Sub 
    WREG = "K"
    GoSub HRSOUT_Sub
    WREG = "E"
    GoSub HRSOUT_Sub
    WREG = "Y"
    GoSub HRSOUT_Sub
    WREG = ":"
    GoSub HRSOUT_Sub
    WREG = " "
    GoSub HRSOUT_Sub
    Conv1Byte_to_2ASCII(KEYNumberByte3)
    Conv1Byte_to_2ASCII(KEYNumberByte2)
    Conv1Byte_to_2ASCII(KEYNumberByte1)
    Conv1Byte_to_2ASCII(KEYNumberByte0)                        
    GoSub HRSOUT#CR_Sub  
    GoSub HRSOUT#CR_Sub

    ' Print the first password at init time.
    WREG = "0"
    GoSub HRSOUT_Sub
    WREG = "0"
    GoSub HRSOUT_Sub
    WREG = "0"
    GoSub HRSOUT_Sub
    WREG = "0"
    GoSub HRSOUT_Sub
    WREG = "0"
    GoSub HRSOUT_Sub
    WREG = " "
    GoSub HRSOUT_Sub
    WREG = "P"
    GoSub HRSOUT_Sub
    WREG = "W"
    GoSub HRSOUT_Sub
    WREG = " "
    GoSub HRSOUT_Sub
    WREG = ":"
    GoSub HRSOUT_Sub 
    WREG = "A"
    GoSub HRSOUT_Sub 
    WREG = "A"
    GoSub HRSOUT_Sub     
    WREG = "A"
    GoSub HRSOUT_Sub 
    WREG = "A"
    GoSub HRSOUT_Sub     
    WREG = "A"
    GoSub HRSOUT_Sub
    WREG = "A"
    GoSub HRSOUT_Sub     
    WREG = "A"
    GoSub HRSOUT_Sub     
    WREG = "A"
    GoSub HRSOUT_Sub         
    GoSub HRSOUT#CR_Sub

    ' Print 1000 incremental passwords
    Clear NumberOfPasswordArray             ' Clear the Passwords counter.
    DataWord = 0
    Repeat
        GoSub RandomPassword_Sub            ' Generate a new password
        GoSub PrintNextPW_Sub               ' Print the new password and counter.
        Inc DataWord
    Until DataWord = 1000
    EEP_EEADR = EEP_ADDR_PASSWORD_ENABLE 
    NVMDAT = $FF                            ' Init the Password system for next reset.
    GoSub Make_EEPROMWrite_Sub
    GoTo WAITING_RESET
$endif
'*******************************************************************************
'* Name    : DECODE AND CHECK THE LINE RECEIVED FROM THE TERMINAL.             *
'*******************************************************************************
START_DECODE:
    #ifdef watchdog_req
    Clrwdt
    #endif
    ' Syncronize the First character.
    HEX32_ChkSum = 0                        ' Clear the CheckSum to start the line. (:) is not included in the CheckSum.
    UARTBuff0_Tail = 1                      ' Point to the second ASCII character of the Buffer0.
    RecordType = 255                        ' Init a different value.
    '===========================================================================
    ' Step 2 - Receive the BYTE COUNT. High Byte + Low Byte (2 ASCII Chars)
    GoSub Read_2AsciiCharacters_Sub         ' Read 2 ASCII bytes and Transform the 2 BCD Bytes to 1 BIN
    NumberHEXBytes = WREG                   ' Number of data HEX Bytes in this line.
    '===========================================================================   
    ' Step 3 - Receive the INICIAL LOW ROM address (High Byte first)
    GoSub Read_2AsciiCharacters_Sub         ' Read 2 ASCII bytes and Transform the 2 BCD Bytes to 1 BIN
    HEX32_ADDRESS.Byte1 = WREG              ' Save the TBLPTRH Byte 

    GoSub Read_2AsciiCharacters_Sub         ' Read 2 ASCII bytes and Transform the 2 BCD Bytes to 1 BIN
    HEX32_ADDRESS.Byte0 = WREG              ' Save the TBLPTRL Byte
    '===========================================================================   
    ' Step 4 - Receive the RECORD TYPE (2 ASCII Chars) High nibble + Low nibble
    GoSub Read_2AsciiCharacters_Sub         ' Read 2 bytes & Transform the 2 ASCII Bytes to 1 BIN
    RecordType = WREG                       ' The Record Type is loaded.
    '=========================================================================== 
    ' Block type 04 - set high address
    ' If the high byte is not zero, an impossible location has been requested - Range Error
    ' If the byte TBLPTRU = 0x01, a new 64KB flash page has been requested - supported for ROM upper than 64KB
    ' If the byte TBLPTRU = 0x20, an ID location has been requested - supported if the option has been enabled.
    ' If the Byte TBLPTRU = 0x30, a CONFIG location has been requested - not supported
    ' If the Byte TBLPTRU = 0xF0, an EEPROM location has been requested- supported.
    ' For a device with <= 64K of flash, any high address other than 0x00F0 is a Range Error
    '===========================================================================
    If RecordType = 4 Then
    '===========================================================================
        ' Step 5 - Receive the HIGH ADDRESS (2 ASCII Chars)
        GoSub Read_2AsciiCharacters_Sub     ' Read 2 ASCII bytes and Transform the 2 BCD Bytes to 1 BIN
        HEX32_ADDRESS.Byte3 = WREG          ' Loading High Address ROM (not used in this program)
        ' If this byte is erroneous the Checksum would do an error?
        '-----------------------------------------------------------------------
        GoSub Read_2AsciiCharacters_Sub     ' Read (2 ASCII Chars) and Transform the 2 BCD Bytes to 1 BIN
        HEX32_ADDRESS.Byte2 = WREG          ' It is the TBLPTRU address value. (= TBLPTRU_HEX32)
        '-----------------------------------------------------------------------
        ' Step 6 - Receive and calculate the Checksum
        ' Read the last (2 ASCII Chars) for checksum.
        GoSub Read2BYTES_CheckSum_Sub
        If F_BootLoader_Error = 1 Then        ' Check the checksum error.
            GoTo SYSTEMERROR
        EndIf
        '-----------------------------------------------------------------------
        ' Write the last download block data of the hex file. (write ROM)
        ' It is not recommended to have hashed code at the end of your code, for the last write block.
        $ifdef _EnableCommand_
        If F_WritingROM = 1 Then
            If EnabledToRead = $55 Then
                If TBLPTRU_HEX32 > 10 Then 
                    If NbDataBytesAvailabled > 0 Then   ' Check if there are some bytes in the buffer.
                        GoSub WRITE_ROMBLOCK_Sub        ' Write the last block.
                        If F_BootLoader_Error = 1 Then
                            GoTo SYSTEMERROR            ' Exit, fatal error.
                        EndIf                       
                    EndIf        
                EndIf
            EndIf
        EndIf
        $else
        If EnabledToRead = $55 Then
            If TBLPTRU_HEX32 > 10 Then 
                If NbDataBytesAvailabled > 0 Then   ' Check if there are some bytes in the buffer.
                    GoSub WRITE_ROMBLOCK_Sub        ' Write the last block.
                    If F_BootLoader_Error = 1 Then
                        GoTo SYSTEMERROR            ' Exit, fatal error.
                    EndIf                       
                EndIf        
            EndIf
        EndIf        
        $endif
        '-----------------------------------------------------------------------
        ' Check if the protocol is correct?
        Select TBLPTRU_HEX32
            Case 0                                  ' YES, Frame OK to begins
                '---------------------------------------------------------------
                ' Check the Device ID
                Read_DeviceID()
                If F_BootLoader_Error = 1 Then    
                    GoTo SYSTEMERROR                ' Exit, fatal error. 
                EndIf
                '---------------------------------------------------------------
                ' Erase the eeprom if this option has been enabled by the user. 
                $ifdef _EnableEraseEeprom_
                EraseEEPROM()
                $endif
                '---------------------------------------------------------------
                ' Erase the FLASH memory.
                EraseROM()
                '---------------------------------------------------------------
                EnabledToRead = $55                 ' Read the datas to write the memory flash.
                '---------------------------------------------------------------                
                UARTBuff1_Head = 0                  ' Clear the UARTBuff1 pointers.
                UARTBuff1_Tail = 0                        
                NbDataBytesAvailabled = 0           ' No data of course.
                NextBlockToWrite = FIRST_BLOCK_TO_WRITE             ' Start write Datas for the first block. ' (Version 3.0)
                TBLPTR_HEX32_PREVIOUS = USER_PROTON_START_ADDRESS   ' (Version 3.0)

            $if _code > FLASH_64K_SIZE_TEST
            Case 1                                  ' ROM Memory Until 128KB, (Case 1,2) => Until 192KB, (Case 1,2,3) => Until 256KB
                EnabledToRead = $55                 ' Read the datas to write the memory flash.
            $endif
                
            Case IDLOC0_Upper 
                ' $20, 32 decimal, an ID location has been requested - supported If the option has been enabled.
                ' Send an information to the terminal.
                $ifdef _EnableWriteIDLOCS_
                EnabledToRead = $3A                     ' Enable to write the ID Locations 
                NbDataBytesAvailabled = 0
                UARTBuff1_Head = 0                      ' Clear the UARTBuff1 pointers.
                UARTBuff1_Tail = 0        
                $else
                EnabledToRead = 0
                $endif

            Case Config1H_Upper
                ' $30, 48 decimal, a CONFIG location has been requested - not supported
                ' Send an information to the terminal.
                EnabledToRead = 0                       ' Nothing to do - not supported.

            $if _eeprom_ > 0
            Case ((EEP_ADDRESS >> 16) & 255)             
                ' $F0, 240 decimal, an EEPROM location has been requested - supported if the option has been enabled.
                ' Send an information to the terminal.
                EnabledToRead = $99                     ' Enable to read data to write the eeprom
                NbDataBytesAvailabled = 0
                UARTBuff1_Head = 0                      ' Clear the UARTBuff1 pointers.
                UARTBuff1_Tail = 0        
            $endif

            $ifdef _EraseEepromCommand_
            Case $50    ' (:020000040050AA) (ASK ME) Special command to erase the eeprom memory.
            EraseEEPROM()
            $endif

            Case Else                                   ' Other Things Not supported: Range Error
                ' An impossible location has been requested - Range Error
                GoTo SYSTEMERROR
        EndSelect    
        '-----------------------------------------------------------------------
        GoTo STARTLINE                                  ' Goto to read more Frames
    '===========================================================================
    ' Block type 00 - incomming the data bytes
    ' The following bytes are data bytes for ROM/EEPROM/IDLOCS
    '===========================================================================
    ElseIf RecordType = 0 Then
    '===========================================================================
        ' The trick of the code
        ' Check if there is an offset for the next line?
        ' DATAs for ROM.
        '-----------------------------------------------------------------------
        If EnabledToRead = $55 Then
            '-------------------------------------------------------------------
            ' Check if the HEX line to write begins at [USER_PROTON_START_ADDRESS]
            If TBLPTR_HEX32 < USER_PROTON_START_ADDRESS Then
                ' The [Declare PROTON_START_ADDRESS = 2048] is forgotten in the USER file.
                GoTo SYSTEMERROR
            EndIf
            '-------------------------------------------------------------------
            $if FLASH_64K_SIZE_TEST > _code
            ' For ROM until 64KB, using 2 bytes address.
            ReadNewBlockToWrite = TBLPTR_HEX32 >> WRITE_BLOCK_SHIFT_BITS   ' Calculate the new Write Block with the new address.
            $else
            ' For ROM upper than 64KB, using 3 bytes address.
            ' Using a Dword variable to make the shift to right.
            DwordNewBlockToWrite.Byte2 = HEX32_ADDRESS.Byte2    ' It is a copy of the TBLPTU of the received data line.
            DwordNewBlockToWrite.Byte1 = HEX32_ADDRESS.Byte1    ' It is a copy of the TBLPTH of the received data Line.
            DwordNewBlockToWrite.Byte0 = HEX32_ADDRESS.Byte0    ' It is a copy of the TBLPTL of the received data Line.
            WREG = WRITE_BLOCK_SHIFT_BITS                       ' Value to decrement the number of shifts.
            ' Shift the address of ROM to the ReadNewBlockToWrite Word variable.

            Set_Bank DwordNewBlockToWrite
            Asm
_SRLBX__255
            Bcf STATUS,0,0
            Rrcf DwordNewBlockToWriteHH,F,0     
            Rrcf DwordNewBlockToWriteH,F,0      
            Rrcf DwordNewBlockToWrite,F,0       
            Decfsz WREG,F,0                     
            Bra _SRLBX__255
            EndAsm
            $endif
            '-------------------------------------------------------------------
            ' Rebuilt the hashed code in the low part of the ROM memory made by the PDS and MPASMWIN. (Version 3.0)
            ' Several Write Blocks could be affected.
            ' It seems to work with 8, 16, 32, 64 Bytes Write Block.
            If NextBlockToWrite = ReadNewBlockToWrite Then
                If TBLPTR_HEX32 > TBLPTR_HEX32_PREVIOUS Then                ' Is there a hole of code?
                    DataWord = TBLPTR_HEX32 - TBLPTR_HEX32_PREVIOUS         ' YES, how bytes?
                    UARTBuff1_Head = UARTBuff1_Head + DataWord.LowByte                 ' Then shift the Head pointer to insert the new line.
                    NbDataBytesAvailabled = NbDataBytesAvailabled + DataWord.LowByte   ' Add this quantity of bytes too.
                EndIf
            EndIf
            TBLPTR_HEX32_PREVIOUS = TBLPTR_HEX32 + NumberHEXBytes        ' Calculate the next line address.
            '-------------------------------------------------------------------
            ' Check if the next line to read is a line for a NO consecutive Write Block.
            If ReadNewBlockToWrite > NextBlockToWrite Then                  ' If it is not the following block... 
                ' Write the incomplete data block before the new offset.
                If NbDataBytesAvailabled > 0 Then                           ' If there are some bytes in the buffer.
                    GoSub WRITE_ROMBLOCK_Sub                                ' Then write them.
                    If F_BootLoader_Error = 1 Then
                        GoTo SYSTEMERROR                                    ' Exit, fatal error.
                    EndIf                       
                    UARTBuff1_Tail = 0                          ' Added Version 2.0
                EndIf
                ' Calculate the offset of the next incomplete block.
                WREG = TBLPTR_HEX32 & MASK_HEAD_ADDRESS_OFFSET  ' Address offset of new data.
                UARTBuff1_Head = WREG                           ' Move ahead the UART Pointer to write the next received byte. 
                NbDataBytesAvailabled = WREG                    ' Then add this offset in bytes to be read.
                NextBlockToWrite = ReadNewBlockToWrite          ' New number of the Block to write.
            EndIf
        EndIf
        '-----------------------------------------------------------------------
        ' Reading all datas of one line.
        Repeat
            ' Step 5 - Extract the Data ASCII Bytes from the UART_Buffer0.
            ' Read 2 ASCII characters (one Binary Byte).
            GoSub Read_2AsciiCharacters_Sub
            GoSub LoadDataToBuffer1_Sub         ' DataBin => Buffer1. YES, then Load the data byte into the buffer
            Inc NbDataBytesAvailabled           ' Increase the number of Bytes received in the buffer.
            Dec NumberHEXBytes
        Until NumberHEXBytes = 0
        '-----------------------------------------------------------------------
        ' Step 6 - Receive and calculate the Checksum Data Bytes received.
        GoSub Read2BYTES_CheckSum_Sub
        If F_BootLoader_Error = 1 Then    ' Check the checksum error.
            GoTo SYSTEMERROR
        EndIf
        '-----------------------------------------------------------------------
        ' Write the full data blocks only.
        If EnabledToRead = $55 Then
            If NbDataBytesAvailabled >= WRITE_FLASH_BLOCKSIZE Then
                UARTBuff1_Tail = 0
                Repeat
                    GoSub WRITE_ROMBLOCK_Sub
                    If F_BootLoader_Error = 1 Then
                        GoTo SYSTEMERROR            ' Exit, fatal error.
                    EndIf
                    $ifdef _EnableCommand_
                    F_WritingROM = 1
                    $endif
                    FullBlock = NbDataBytesAvailabled / WRITE_FLASH_BLOCKSIZE
                Until FullBlock = 0                 ' Write full blocks only.
                If NbDataBytesAvailabled = 0 Then
                    UARTBuff1_Head = 0              ' Clear the pointers to receive a new Block.
                    UARTBuff1_Tail = 0          
                EndIf
            EndIf
        EndIf
        '-----------------------------------------------------------------------
        ' Write the Data to the ID Locations (once).
        $ifdef _EnableWriteIDLOCS_
        If EnabledToRead = $3A Then
            If NbDataBytesAvailabled > 0 Then        
                GoSub WRITE_IDLOCS_Sub          ' For IDLOC 
                If F_BootLoader_Error = 1 Then    
                    GoTo SYSTEMERROR            ' Exit, fatal error. 
                EndIf
            EndIf
        EndIf
        $endif
        '-----------------------------------------------------------------------
        $if _eeprom_ > 0
        ' Write the Data to the eeprom memory.
        If EnabledToRead = $99 Then             ' Enable to read the buffer with code.
            If NbDataBytesAvailabled > 0 Then
                GoSub WRITE_EEPLine_Sub         ' For eeprom it is more easy to write line by line.
                If F_BootLoader_Error = 1 Then
                    GoTo SYSTEMERROR            ' Exit, fatal error.
                EndIf
                UARTBuff1_Head = 0              ' Clear the UARTBuff1 pointers in every line.
                UARTBuff1_Tail = 0        
            EndIf 
        EndIf
        $endif
        '-----------------------------------------------------------------------
        GoTo STARTLINE                          ' Goto to read more lines
    '===========================================================================
    ' Block type 01 - Final line sequence of the HEX file.
    '===========================================================================
    ElseIf RecordType = 1 Then
    '===========================================================================
        ' Receiving the Final sequence.
        ' Step 7 - Receive and calculate the Checksum
        GoSub Read2BYTES_CheckSum_Sub
        If F_BootLoader_Error = 1 Then
            GoTo SYSTEMERROR                    ' Exit, CheckSumFrame fatal error.
        EndIf        
        '-----------------------------------------------------------------------
        If NumberHEXBytes = 0 Then
            ' The final line is correct, now write the "Goto PIC#Loader#Boot" in the position 0 of ROM memory. 
            '*******************************************************************
            '* Name    : Generate a new PassWord                               *
            '* Purpose : For the next download                                 *
            '*******************************************************************
            $ifdef _EnablePassWordMode0_
            GoSub RandomPassword_Sub
            $endif
            '*******************************************************************
            GoTo RUNUSERCODE                    ' Starts the USER CODE.
        EndIf
    EndIf
    '===========================================================================
    ' All others conditions are an error only.
    '===========================================================================
    GoTo SYSTEMERROR
'===============================================================================
'=== MAIN LOADER PROGRAM END ===================================================
'===============================================================================
'---SUBROUTINES:----------------------------------------------------------------
'===============================================================================
'*******************************************************************************
'* Name    : Read_2AsciiCharacters_Sub                                         *
'* Purpose : Read a group of 2 chararcters with CheckSum                       *
'* Output  : WREG, DataBin                                                     *
'*******************************************************************************
Read_2AsciiCharacters_Sub:
    ' Read the first ASCII character & return a HEX byte. (HEX high niblle)
    GoSub ReadBuffer0_Sub
    DataWord.HighByte = WREG

    ' Read the second ASCII character & Return a HEX Byte. (HEX Low niblle)
    GoSub ReadBuffer0_Sub
    DataWord.LowByte = WREG
'===============================================================================
' Transform the 2 HEX Bytes to 1 BIN 
'===============================================================================
TWOHEX_1BIN_Sub:
    WREG = DataWord.HighByte << 4
    WREG = DataWord.LowByte | WREG
    ' DO the checksum: Using the WREG value.
    DataBin = WREG
    If F_EnableReadCKS = 0 Then
        HEX32_ChkSum = HEX32_ChkSum + WREG
    EndIf
    F_EnableReadCKS = 0             ' Clear the flag for next time.
'    WREG = DataBin                  ' Load the WREG for next code. (comment the line = save one asm line)
    Return
'*******************************************************************************
'* Name    : LoadDataToBuffer1_Sub                                             *
'* Purpose : Load data address into buffer1                                    *
'*******************************************************************************
LoadDataToBuffer1_Sub:
    UART_Buffer1[UARTBuff1_Head] = DataBin  ' Load a data address byte into the buffer1
    Inc UARTBuff1_Head                      ' array pointer for next data byte  
    Return
'*******************************************************************************
'* Name    : SYSTEMERROR                                                       *
'* Purpose : Management of the errors of the Bootloader                        *
'*         : Error Messages are sent to the terminal.                          *
'*******************************************************************************
SYSTEMERROR:
    $ifdef _EnableMessages_
    GoSub SendErrorMessage_Sub  ' Send an error message
    $endif
    '===========================================================================
    ' Because an error ocurred then goto receiving bytes from computer until the HEX file closes.
    '===========================================================================
    F_StopDecodingLines = 1     ' The bootloader will receive bytes only.
    GoTo STARTLINE
'*******************************************************************************
'* Name    : Read2BYTES_CheckSum_Sub                                           *
'* Purpose : Transform 2 bytes HEX to 1 Byte BIN                               *
'*******************************************************************************
Read2BYTES_CheckSum_Sub:
    F_EnableReadCKS = 1                 ' Disable calculate the checksum
    GoSub Read_2AsciiCharacters_Sub     ' Receive 2 ASCII characters or one binary byte
    CheckSumFrame = WREG                ' Loading the line CheckSumFrame    
'===============================================================================
' Make the Check summ: The 2 values are not destroyed.
'===============================================================================
    WREG = HEX32_ChkSum
    WREG = CheckSumFrame + WREG
    If WREG > 0 Then
        F_BootLoader_Error = 1
    EndIf
    '---------------------------------------------------------------------------
    Return
'*******************************************************************************
'* Name    : ReadBuffer0_Sub                                                   *
'* Purpose : Take an ASCII byte from the Buffer0, transform to Binary          *  
'*           Return ASCII in DataByte, Return Binary in WREG                   *
'*******************************************************************************
ReadBuffer0_Sub:
    DataByte = UART_Buffer0[UARTBuff0_Tail]
    Inc UARTBuff0_Tail

ASCII_to_BCD_Sub:
    ' Transform the ASCII to BCD in DataByte and return the value in WREG.
    If DataByte > 64 Then      
        WREG = DataByte - 55
    Else
        WREG = DataByte - 48
    EndIf
    ' The BIN character is in WREG & the ASCII character is in DataByte 
    Return
'*******************************************************************************
'* Name    : CleanUARTErrors_Sub                                               *
'* Purpose : Clean UART errors for UART 1 Or UART2.                            *
'*******************************************************************************
CleanUARTErrors_Sub:
    Clear _CREN_        ' Clear continuous receive bit
    Set _CREN_          ' and set it again  
    WREG = _RCREG_      ' Read the Byte and discard
    _TXSTA_.6 = 1       ' Nine bit mode for 2 stop bits for transmision. (V2.1)
    _TXSTA_.0 = 1       ' Send the first stop bit. The second stop bit is set by the UART. (V2.1)
    Return 
'*******************************************************************************
'* Name    : HRSOUT#MSTR_Sub                                                   *
'* Purpose : Send a CData String to the UART. Loading Label: TBLPTRW = TEXT2    *
'*******************************************************************************
' Send a string to the terminal without CR or LF.
' A copy of the HRSout command of the PDS.
'HRSOUT#MSTR_Sub:
'    $if FLASH_64K_SIZE_TEST > _code
'    TBLPTRU = 0                 ' (Modified V1.1)
'    $endif
'    While 1 = 1
'        Tblrd*+    
'        If TABLAT = 0 Then Return
'        WREG = TABLAT
'        GoSub HRSOUT_Sub        ' Send the ASCII character to the terminal.
'    Wend
'-------------------------------------------------------------------------------
' Send a string to the terminal with CR & LF.
'HRSOUT#MSTR_CRLF_Sub:
'    GoSub HRSOUT#MSTR_Sub
'*******************************************************************************
'* Name    : SendErrorMessage_Sub                                              *
'* Purpose : Send the "E" error message individually + CR + LF                 *
'*******************************************************************************
    $ifdef _EnableMessages_
SendErrorMessage_Sub:
    WREG = "E"
    GoSub HRSOUT_Sub
    $endif
'*******************************************************************************
'* Name    : HRSOUT#CR_Sub                                                     *
'* Purpose : Send CR + LF To the terminal.                                     *
'*******************************************************************************
HRSOUT#CR_Sub:
    WREG = CR
    GoSub HRSOUT_Sub
    WREG = LF               ' Line Feed could be necessary to avoid an issue with the terminal.
'    GoSub HRSOUT_Sub
'    Return
'*******************************************************************************
'* Name    : HRSOUT_Sub                                                        *
'* Purpose : Send a Byte To the UART 1 or UART2 with 2 stop bits.              *
'*******************************************************************************
HRSOUT_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif 

    ' This code take care of the different Banks of TXxSTA.
    ' Waiting UART ready.
    While 1 = 1
        PRODH = _TXSTA_         ' DataByte is a Byte System.
        If PRODH.1 = 1 Then Break
    Wend

    ' To avoid some old PIC's errata the bits are set after sent the last byte.
    _TXSTA_.6 = 1               ' Nine bit mode for 2 stop bits for transmision.
    _TXSTA_.0 = 1               ' Send the first stop bit. The second stop bit is set by the UART.
    _TXREG_ = WREG              ' Start sending the byte.
    Return   
'*******************************************************************************
'* Name    : EEPROMWrite                                                       *
'* Purpose : Write a single byte of data to on board EEPROM                    *
'*         : The data is in WREG and the address in EEP_EEADR = NVMDAT         *
'*******************************************************************************
$if _eeprom_ > 0
Make_EEPROMWrite_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif
    NVMCON1 = %00000100         ' Setup write eeprom
    GoTo  WriteSequence_Sub     ' Required write sequence
$endif
'*******************************************************************************
'* Name    : Make_EraseBlockFLASH_Sub                                          *
'* Purpose : Confirm To erase one block of the  FLASH memory                   *
'*******************************************************************************
Make_EraseBlockFLASH_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif
    NVMCON1 = %10010100         ' Flash acces + Start erase flash + Setup write flash 
    GoTo  WriteSequence_Sub     ' Required write sequence 
'*******************************************************************************
'* Name    : Make_EraseBlockConfig_Sub                                         *
'* Purpose : Confirm To erase one block of the  FLASH memory                   *
'*******************************************************************************
$ifdef _EnableWriteIDLOCS_
Make_EraseBlockConfig_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif
    NVMCON1 = %11010100         ' Flash acces + Start erase flash + Setup write flash 
    GoTo  WriteSequence_Sub     ' Required write sequence 
$endif
'*******************************************************************************
'* Name    : Make_WriteCONFIG_Sub                                              *
'* Purpose : Confirm To write the data To CONFIG memory                        *
'*         : ID LOCS are same as Config Fuses (errata PIC18F27K40 Version A2)  * 
'*******************************************************************************
$ifdef _EnableWriteIDLOCS_
Make_WriteCONFIG_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif
    NVMCON1 = %11000100         ' Setup write flash
    GoTo  WriteSequence_Sub     ' Required write sequence 
$endif
'*******************************************************************************
'* Name    : Make_WriteFLASH_Sub                                               *
'* Purpose : Confirm To write the data To FLASH memory                         *
'*******************************************************************************
Make_WriteFLASH_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif
    NVMCON1 = %10000100         ' Setup write flash  
'*******************************************************************************
'* Name    : WriteSequence                                                     *
'* Purpose : A required write sequence for EEPROM and FLASH writes             *
'*******************************************************************************
WriteSequence_Sub:
    NVMCON2 = $55                    ' EECON2 <- $55
    NVMCON2 = $AA                    ' EECON2 <- $AA
    Set NVMCON1Bits_WR               ' Set WR Bit For write
    Nop
'    If NVMCON1Bits_WRERR = 1 Then   ' Checking the NVMCON1Bits_WRERR
'        F_BootLoader_Error = 1
'    Endif
    While NVMCON1Bits_WR = 1 : Wend  ' Waiting to complete.
    NVMCON1 = 0                      ' Disable write
    Return
'*******************************************************************************
' Name    : ReadDeviceID_Sub                                                   *
' Purpose : Read a single byte from ROM                                        *
' Input   : none                                                               *
' Output  : PICPartNumber & (PICRevision is disable to save ROM space)         *
'*******************************************************************************
'3.6 Device ID and Revision ID
'    The 16-bit device ID word is located at 3F FFFEh and the 16-bit revision ID is
'    located at 3F FFFCh. These locations are read-only and cannot be erased or modified.
' In asm file:
'    DEVID1 equ 0X3FFFFE
'    DEVID2 equ 0X3FFFFF
'
' Revision ID address. (3FFFFCh-3FFFFFFh) Location of (DEVID1, DEVID2) in memory map.
' see the datasheet
'------------------------------------------------------------------------------- 
ReadDeviceID_Sub:
    ' Read the DEVID1 (Device ID) from the PPI file.
    NVMCON1 = %11000000              ' Access Configuration registers 
    Set_Bank TBLPTRL
    Asm
    Movlw Low (DEVID1-2)
    Movwf TBLPTRL
    Movlw High (DEVID1-2)
    Movwf TBLPTRH
    Movlw upper (DEVID1-2)
    Movwf TBLPTRU
    EndAsm
    '
    $ifdef _EnablePrintRevisionID_
    ' Read Revision
    Tblrd*+
    PICRevision.LowByte = TABLAT
    Tblrd*+
    PICRevision.HighByte = TABLAT     
    $else
    Tblrd*+
    Tblrd*+    
    $endif
    '
    ' Read Part Number
    Tblrd*+
    PICPartNumber.LowByte = TABLAT
    Tblrd*+
    PICPartNumber.HighByte = TABLAT
    NVMCON1 = 0                     ' Disable Access the Configuration registers
    TBLPTRU = 0
    '---------------------------------------------------------------------------
    $ifdef _EnablePrintRevisionID_
    ' Calculate the PIC Part Number & revision
    ' ASCII calculation
    PICRevision.HighByte = PICRevision.HighByte & %00001111
    PICRevision.HighByte = PICRevision.HighByte << 2
    DataByte = PICRevision.LowByte >> 6
    DataByte = PICRevision.HighByte | DataByte
    DataByte = DataByte + "A"
    ' Number calculation
    PICRevision.LowByte = PICRevision.LowByte & %00111111
    $endif
'===============================================================================
' OPTION: Print PIC Part Number & the Device ID to the terminal
'===============================================================================
    $ifdef _EnablePrintDeviceID_
    ' PRINT the Part Number
    WREG = "I"
    GoSub HRSOUT_Sub
    WREG = "D"
    GoSub HRSOUT_Sub
    WREG = "$"
    GoSub HRSOUT_Sub        
    Conv1Byte_to_2ASCII(PICPartNumber.HighByte)
    Conv1Byte_to_2ASCII(PICPartNumber.LowByte)
    '
    $ifdef _EnablePrintRevisionID_
    ' PRINT the Chip revision
    WREG = " "
    GoSub HRSOUT_Sub
    WREG = "R"
    GoSub HRSOUT_Sub
    WREG = DataByte
    GoSub HRSOUT_Sub
    WREG = "$"
    GoSub HRSOUT_Sub
    Conv1Byte_to_2ASCII(PICRevision.LowByte)
    $endif
    '
    GoSub HRSOUT#CR_Sub
    $endif
'===============================================================================
' Compare the PIC ID with the user ID
'===============================================================================
    If PICPartNumber <> DEVICE_ID Then
        F_BootLoader_Error = 1                    ' "DIF" - Device ID Fails - PIC not correct.
    EndIf
    '---------------------------------------------------------------------------
    Return    
'*******************************************************************************
'* Name    : Erase all the EEPROM in the RANGE                                 *
'* Purpose : Write a $FF byte in all addresses of the eeprom memory            *
'*         : The code takes care of the dimension of the eeprom memory         *
'*******************************************************************************
$if _defined(_EnableEraseEeprom_) Or _defined(_EraseEepromCommand_)
EraseEEPROM_Sub:
    '---------------------------------------------------------------------------
    LOOP_EEADR = EEP_ERASE_LOW_ADDRESS
    Repeat
        EEP_EEADR = LOOP_EEADR
        NVMDAT = $FF                ' NVMDAT <= 0xFF is the erased value
        GoSub Make_EEPROMWrite_Sub  ' Write the character to the eeprom.    
        Inc LOOP_EEADR
    Until LOOP_EEADR > EEP_ERASE_ADDRESS_MAX
    '---------------------------------------------------------------------------
    Return
$endif
'*******************************************************************************
'* Name    : EraseUserROM_Sub                                                  *
'* Purpose : Erase all ROM Memory until Bootloader origin block                *
'*******************************************************************************
EraseUserROM_Sub:
'    F_BlocksErased = 1                  ' Write a flag to know that the ROM has been erased.
    BlockToErase = FIRST_BLOCK_TO_ERASE ' First block to erase. (version 3.0)
    '---------------------------------------------------------------------------
    Repeat
        ' Shifting directly the BlockToErase into TBLPTRU, TBLPTRH, TBLPTRL. The Same result but with less lines code.
        TBLPTRH = BlockToErase.HighByte
        TBLPTRL = BlockToErase.LowByte
        LOOPByte = ERASE_BLOCK_SHIFT_BITS
        GoSub Shift_TBLPTR_To_Left_Sub
        ' The result is the ROM address of the first byte of the Block to ERASE.
        '=======================================================================
        GoSub Make_EraseBlockFLASH_Sub    ' Setup writes, Start erase block, required write sequence, Disable writes
        '=======================================================================
        Inc BlockToErase
    Until BlockToErase > LAST_BLOCK_TO_ERASE  ' (version 3.0)
    '---------------------------------------------------------------------------
    Return
'*******************************************************************************
'* Name    : Some subroutines                                                  *
'* Purpose : To save lines of code                                             *
'*******************************************************************************
READ_UARTBuff1_to_WREG:
    WREG =  UART_Buffer1[UARTBuff1_Tail]  ' Buffer Data => WREG
    Return
'*******************************************************************************
'* Name    : WRITE_EEPLine_Sub                                                 *
'* Purpose : Write the eeprom line by line                                     *
'*******************************************************************************
$if _eeprom_ > 0
WRITE_EEPLine_Sub:
'===============================================================================  
' Load the eeprom Address from the HEX file.
'===============================================================================  
    EEP_EEADR = TBLPTR_HEX32    ' Address of EEPROM to write this line.                
'=============================================================================== 
' Write the eeprom memory.
'=============================================================================== 
WriteEEP_ReadBuffer:
    Repeat
        '-----------------------------------------------------------------------
        ' Check the address of the eeprom to write: Protecting the Password area.
        $ifdef _EnablePassWordMode0_
        If EEP_EEADR > (EEP_ADDR_PASSWORD_SYSTEM - 1) Then
            F_BootLoader_Error = 1          ' Error to write the eeprom.
            Return                          ' Return a fatal error.        
        EndIf
        $endif
        '-----------------------------------------------------------------------
        GoSub READ_UARTBuff1_to_WREG        ' WREG <= Buffer Data
        NVMDAT = WREG
        DataByte = WREG                     ' Copy the result.
        GoSub Make_EEPROMWrite_Sub          ' WREG => eeprom
        'Read the eeprom to check the data
        ''NVMCON1 = 0                          ' Acces to eeprom Data.
        Set NVMCON1bits_RD                  ' EEProm read (the NVMCON1 is already = 0) 
        '-----------------------------------------------------------------------
        ' Check the NVMDAT if correct?
        If NVMDAT <> DataByte Then          ' Check if the byte written in ROM is valid?
            F_BootLoader_Error = 1          ' "WEF" - "Write Eeprom Fails", Error to write the eeprom.
            Return                          ' Return a fatal error.
        EndIf
        Inc UARTBuff1_Tail                  ' Next byte to extract from the Buffer.
        Inc EEP_EEADR                       ' Next EEP address to write
        Dec NbDataBytesAvailabled           ' Number Bytes received ready to write the EEPROM.
    Until NbDataBytesAvailabled = 0 
    ' Eeeprom already written and checked. All eeprom bytes OK.
    GoSub ClearUARTBuffer1_Sub             ' Clear the buffer again because it has been altered. (Load all positions with $FF)
    '---------------------------------------------------------------------------
    Return
    $endif
'*******************************************************************************
'* Name    : WRITE_ROMBLOCK_Sub                                                *
'* Purpose : Write data to a WRITE_FLASH_BLOCKSIZE                             *
'*******************************************************************************
WRITE_ROMBLOCK_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif
'===============================================================================
' Check if the code tries To write into flashloader segment Or higher
'===============================================================================
    If NextBlockToWrite >= FIRST_BLOCK_TO_WRITE Then GoTo ContinueToWrite
    If NextBlockToWrite <= LAST_BLOCK_TO_WRITE Then GoTo ContinueToWrite
    F_BootLoader_Error = 1                  ' Return a fatal error (Out of Range)
    Return
ContinueToWrite:
'===============================================================================
' Load the data into the write buffer.  (preview)
'===============================================================================
    F_WritingFlash = 1      ' Flag to control the error file.
WRITEROM:
    UARTBuff1_Tail_COP = UARTBuff1_Tail     ' Save the Tail pointer of buffer1.
    GoSub Load_TBLPTR_BlockAddress_Sub      ' Load the first address of the block (NextBlockToWrite) in TBLPTRL/TBLPTRH/TBLPTRU
    LOOPByte = 0
    NVMCON1 = 128                           ' Access Flash memory
    Repeat
        If NbDataBytesAvailabled > 0 Then
            GoSub READ_UARTBuff1_to_WREG    ' WREG <- Buffer Data
            TABLAT = WREG
            Dec NbDataBytesAvailabled       ' Number Bytes received ready to write the ROM
        Else
            TABLAT = $FF        
        EndIf
        Asm
        tblwt*+                             ' Write to holding register 
        EndAsm
        Inc UARTBuff1_Tail                  ' Increment the reading of the buffer.
        Inc LOOPByte                        ' Number bytes to extract.
    Until LOOPByte = WRITE_FLASH_BLOCKSIZE 
'===============================================================================
' Flash the ROM WRITE BLOCKSIZE
'===============================================================================
    Tblrd*-                                 ' point back to the last block, need this before write...
    GoSub Make_WriteFLASH_Sub               ' Setup writes + write sequence + Disable writes
'===============================================================================
VERIFYROM:
' Read and compare the values of the buffer and the ROM
'===============================================================================
    UARTBuff1_Tail = UARTBuff1_Tail_COP         ' The pointer has been used, restore it.
    GoSub Load_TBLPTR_BlockAddress_Sub          ' Load the first address of the block
    LOOPByte = 0
    NVMCON1 = 128                               ' Access Flash memory
    Repeat
        Tblrd*+                                 ' Read one memory position & increment ROM pointer.
        DataByte2 = TABLAT
        '-----------------------------------------------------------------------
        GoSub READ_UARTBuff1_to_WREG            ' Read every byte from the buffer
        If WREG <> DataByte2 Then               ' Check if the byte written in ROM is valid?
            F_BootLoader_Error = 1              ' "GSF" - No Good - hex32 record fails Or has other flaws
            Return                              ' Exit, fatal error.
        EndIf
        '-----------------------------------------------------------------------
        UART_Buffer1[UARTBuff1_Tail] = $FF      ' $FF => Buffer Data. The byte written is correct. Load a NOP with the $FF value.
        Inc UARTBuff1_Tail                      ' Increment the reading of the buffer.
        Inc LOOPByte                            ' The UARTBuff1_Tail point to the next byte to write. 
    Until LOOPByte = WRITE_FLASH_BLOCKSIZE      ' All bytes of this block are valid.    
    '---------------------------------------------------------------------------    
    Inc NextBlockToWrite    ' Pointing to the next block to write.
    Return
'*******************************************************************************
'* Name    : WRITE_IDLOCS_Sub                                                  *
'* Purpose : Write INFO to the ID LOCATIONS                                    *
'*******************************************************************************
$ifdef _EnableWriteIDLOCS_
WRITE_IDLOCS_Sub:
' The bits NVMREG1 & NVMREG0 of NVMCON1 must be set to acces the User ID.
'===============================================================================
' Load the address of ID_LOCATION & ERASE the ID Locations block first
'===============================================================================
    GoSub Load_IDLOC03_Sub              ' Load Address of the ID Locations version 4.2
'    TBLPTRL = IDLOC0_Low
'    TBLPTRH = IDLOC0_High
'    TBLPTRU = IDLOC0_Upper
    GoSub Make_EraseBlockConfig_Sub     ' Setup write, Start erase block, ERASE CONFIG BLOCK, required write sequence, Disable writes
'===============================================================================
' Write the data of the ID Locations (preview)
'===============================================================================
    UARTBuff1_Tail = 0                  ' Clear the UARTBuff1 read pointer.
    NVMCON1 = 192                       ' %11000000, Access Config memory
    Repeat
        GoSub READ_UARTBuff1_to_WREG    ' WREG <- Buffer Data
        TABLAT = WREG                   ' Load TABLAT
        Asm
        tblwt*+                         ' write to holding register 
        EndAsm
        Inc UARTBuff1_Tail              ' Next byte to extract from the Buffer.
'    Until UARTBuff1_Tail > 7
    Until UARTBuff1_Tail > NbDataBytesAvailabled
'===============================================================================
' Write the ID Locations, Flash the ROM.
'===============================================================================
    GoSub Load_IDLOC02_Sub              ' Load Address of the first ID Locations version 4.2
'    TBLPTRL = IDLOC0_Low                ' Load Address of the first ID Locations version 4.2
'    TBLPTRH = IDLOC0_High               ' Load Address of the first ID Locations version 4.2
    GoSub Make_WriteCONFIG_Sub          ' Setup writes + write sequence + Disable writes
'    If F_BootLoader_Error = 1 Then      ' Check the NVMCON1Bits_WRERR
'        Return
'    EndIf
'===============================================================================
' Verifiy the IDLOCS in ROM & compare with the buffer.
'===============================================================================
    GoSub Load_IDLOC02_Sub              ' Load Address of the first ID Locations version 4.2
'    TBLPTRL = IDLOC0_Low                ' Load Address of the first ID Locations version 4.2
'    TBLPTRH = IDLOC0_High               ' Load Address of the first ID Locations version 4.2
    UARTBuff1_Tail = 0                  ' Clear the UARTBuff1 read pointer.
    NVMCON1 = 192                       ' %11000000, Access Config memory
    Repeat                              ' Start the loop
        Tblrd*+                         ' Read one memory position.
        GoSub READ_UARTBuff1_to_WREG    ' Buffer Data => WREG 
        If WREG <> TABLAT Then          ' Compare the buffer with the ROM.
            F_BootLoader_Error = 1      ' If error then exit.
            Return                      ' Exit, fatal error.
        EndIf
        Inc UARTBuff1_Tail              ' The UARTBuff1_Tail point to the next byte to read. 
'    Until UARTBuff1_Tail > 7
    Until UARTBuff1_Tail > NbDataBytesAvailabled
    NVMCON1 = 0
    GoSub ClearUARTBuffer1_Sub          ' Buffer: $FF => Data. Clear the buffer again because it has been altered. (Load all positions with $FF)
    Return
$endif
'*******************************************************************************
'* Name    : Load_TBLPTR_BlockAddress_Sub                                      *
'* Purpose : Shifting the NextBlockToWrite Word variable to TBLPTRW/TBLPTRU     *
'*         : TBLPTRW/TBLPTU = NextBlockToWrite << WRITE_BLOCK_SHIFT_BITS        *
'*******************************************************************************
Load_TBLPTR_BlockAddress_Sub:
    ' Point the ROM pointer to the first address of the write block.
    TBLPTRH = NextBlockToWrite.HighByte
    TBLPTRL = NextBlockToWrite.LowByte
    LOOPByte = WRITE_BLOCK_SHIFT_BITS
'    GoSub Shift_TBLPTR_To_Left_Sub
'    Return
'*******************************************************************************
'* Name    : Shift_TBLPTR_To_Left_Sub                                          *
'* Purpose : Clear TBLPTRL & TBLPTRH & TBLPTRU                                 *
'*******************************************************************************
Shift_TBLPTR_To_Left_Sub:
    TBLPTRU = 0
    WREG = LOOPByte
$if FLASH_64K_SIZE_TEST < _code
    ' ROM upper than 64KB
    Set_Bank TBLPTRL
    Asm
_SLLBX__5555
    Bcf STATUS,0,0
    Rlcf TBLPTRL,F,0
    Rlcf TBLPTRH,F,0
    Rlcf TBLPTRU,F,0
    Decfsz WREG,F,0
    Bra _SLLBX__5555
    Movlb 0
    EndAsm
$else
    ' ROM less or equal than 64KB
    Set_Bank TBLPTRL
    Asm
_SLLBX__5555
    Bcf STATUS,0,0
    Rlcf TBLPTRL,F,0
    Rlcf TBLPTRH,F,0
    Decfsz WREG,F,0
    Bra _SLLBX__5555
    Movlb 0
    EndAsm    
$endif
    Return
'*******************************************************************************
'* Name    : Load_IDLOC03_Sub/Load_IDLOC02_Sub                                 *
'* Purpose : Load IDLOC0 address into TBLPTRL & TBLPTRH & TBLPTRU registers    *
'*******************************************************************************
Load_IDLOC03_Sub:
    TBLPTRU = IDLOC0_Upper
Load_IDLOC02_Sub:
    TBLPTRH = IDLOC0_High
    TBLPTRL = IDLOC0_Low
    Return
'*******************************************************************************
'* Name    : Clear_TBLPTR_Sub                                                  *
'* Purpose : Clear TBLPTRL & TBLPTRH & TBLPTRU                                 *
'*******************************************************************************
'Clear_TBLPTR_Sub:
'    TBLPTRL = 0   
'Clear_TBLPTR2_Sub:
'    TBLPTRH = 0
'    TBLPTRU = 0
'    Return     
'*******************************************************************************
'* Name    : ClearUARTBuffer1_Sub                                              *
'* Purpose : Fill the UART buffer with Nop's ($FF)                             *
'*******************************************************************************
ClearUARTBuffer1_Sub:
    Set UART_Buffer1
    Return         
'*******************************************************************************
'* Name    : CONVERT_LOWNIBBLE_to_ASCII_Sub (Conv_LoNib_to_ASCII(pValue))      *
'* Purpose : Convert the a low nibble to ASCII character                       *
'* Input   : Data in WREG                                                      *
'* Output  : Data out in WREG                                                  *
'*******************************************************************************
$if _defined(_EnablePrintDeviceID_) Or _defined(_EnablePassWordMode0_) Or _defined(_EnablePassWordMode1_)
CONVERT_LOWNIBBLE_to_ASCII_Sub:
    WREG = WREG & $0F   ' Take the lower nibble only.
    Asm
    Addlw   -10         ; Test to see if W < 10
    Btfsc   STATUS,0    ; If yes, skip ahead
    Addlw   "A"-"0"-10  ; Add ASCII 'A', and negate next line
    Addlw   "0"+10      ; Add ASCII '0' and original 10
    EndAsm
    Return   
$endif
'*******************************************************************************
'* Name    : CONV_1BIN_2HEX_Sub                                                *
'* Purpose : Convert 1 Binary Byte to 2 HEX Bytes                              *
'*******************************************************************************
' Conv1Byte_to_2ASCII_CRLF(Byte)
' Send a Byte Bin to the terminal in HEX ASCII with sending a CR,LF
'CONV_1BYTE_2HEX_Sub:
'    WREG = Byte_DSB >> 4        ' Convert the high nibble.
'    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub
'    GoSub HRSOUT_Sub
'    WREG = Byte_DSB             ' Convert the low nibble.
'    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub
'    GoSub HRSOUT_Sub
'    GoTo HRSOUT#CR_Sub          ' Send a CR,LF
'===============================================================================
' Send a Byte Bin to the terminal in HEX ASCII without sending CR,LF
' Conv1Byte_to_2ASCII(Byte)
'===============================================================================
$if _defined(_EnablePrintDeviceID_) Or _defined(_EnableToPrintCodes_)
CONV_1BYTE_2HEXBB_Sub:
    WREG = Byte_DSB >> 4        ' Convert the high nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub
    GoSub HRSOUT_Sub

CONV_1BYTE_1HEXBB11_Sub:
    WREG = Byte_DSB             ' Convert the low nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub
    GoTo HRSOUT_Sub
$endif
'*******************************************************************************
'* Name    : UART TIMEOUT: TESTING THE FLASH MEMORY FIRST BLCOK                *
'* Purpose : After an UART Time Out, Close or goto user code or standby        *
'*******************************************************************************
TIMEOUT00:
    $ifdef _EnableMessages_
    GoSub SendErrorMessage_Sub          ' Send an error message (E).
    $endif
    GoTo WAITING_RESET

TIMEOUT:
' Testing if any code has been flashed in memory.
    #ifdef watchdog_req
    Clrwdt
    #endif
    TBLPTRU = 0 
    TBLPTRW = ADDRESS_USER_CODE_CHECK   ' (2080, $0820) Address In the first write block of user memory.
    NVMCON1 = 128                       ' Access Flash memory
    Tblrd*+
    WREG = TABLAT
    Tblrd*+
    WREG = TABLAT ^ WREG
    Tblrd*+
    WREG = TABLAT ^ WREG
    NVMCON1 = 0
    '---------------------------------------------------------------------------
    If WREG = $FF Then GoTo WAITING_RESET           ' No code has been flashed or ROM already erased
''    If F_BlocksErased = 1 Then GoTo WAITING_RESET   ' No code has been flashed or ROM already erased
    '---------------------------------------------------------------------------
    ' There is some code there...
'*******************************************************************************
'* Name    : RunUserCode                                                       *
'* Purpose : Close the bootloader, goto user code or standby                   *
'*******************************************************************************
RUNUSERCODE:    ' (Version 3.0)     ' The upload ran with succes, goto user code.
    XON()                           ' Change the terminal to normal state.
'===============================================================================
' DO not modify: automatic settings.
'===============================================================================
    _TXCON_ = 0         ' Stop the generic Timer.
    _TXSTA_ = 0         ' Reset transmit STATUS And control register
    _RCSTA_ = 0         ' Reset receive STATUS And control register
    $ifdef _BAUDCON_
    _BAUDCON_ = 0
    $endif

''    $ifdef _Config1_Pins_SFR_
''    UART_ANSEL = 255        ' All bits to analog
''    $endif
    Input UART_PORT         ' All bits used changed for inputs

    TBLPTRU = 0
    STKPTR = 0              ' Reset the stack pointer 
    GoTo ProtonCodeStart    ' Goto the user code (ProtonCodeStart) (Org 2048) (version 4.0)
'===============================================================================
'===============================================================================
WAITING_RESET_TIMEOUT_ERR:
    ' TIMEOUT after an error.
    ' Version 3.1 More security. If error then erase the block of the ADDRESS_USER_CODE_CHECK address.
    ' Erase the second block of ROM if Error & TimeOut.
    If F_WritingFlash = 1 Then
        TBLPTRU = 0  
        TBLPTRW = ADDRESS_USER_CODE_CHECK & MASK_ERASE_BLOCK    ' Set the address of the block to be erased.
        GoSub Make_EraseBlockFLASH_Sub                         ' Setup write, Start erase block, ERASE_BLOCK, required write sequence, Disable writes
    EndIf
    '===========================================================================
WAITING_RESET:
    $ifdef _EnableMessages_
    WREG = "W"
    GoSub HRSOUT_Sub
    GoSub HRSOUT#CR_Sub
    $endif
    XON()                       ' Change the terminal to normal state.
'-------------------------------------------------------------------------------
WAITING_RESET44:
    While 1 = 1                 ' This code is needed when the Watchdog is ON (always enabled) in Config Fuses.
        #ifdef watchdog_req
        Clrwdt
        #endif
    Wend                        
'*******************************************************************************
'* Name    : PASSWORD generation (RANDOM)                                      *
'* Purpose : Every time the Bootloader writes to ROM memory, Password changes  *
'*******************************************************************************
$ifdef _EnablePassWordMode0_
RandomPassword_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif
    GoSub IncBCDCounters_Sub                ' Increment the counter of Passwords.
    ' To read the eeprom only
    EEP_EEADR = EEP_ADDR_PASSWORD_ENABLE            
    NVMCON1 = 0
    NVMCON1Bits_RD = 1

    $ifndef _EnableToPrintCodes_
    '---------------------------------------------------------------------------
    ' Writing 1000 times to the eeprom is not allowed.
    If NVMDAT <> $A5 Then                  ' Is it the first time?
        '=======================================================================
        ' LOAD THE ORIGINAL KEY AT POWER UP: (Reset Password)
        '=======================================================================
        SHIFT_REG3 = KEYNumberByte3 
        SHIFT_REG2 = KEYNumberByte2    
        SHIFT_REG1 = KEYNumberByte1
        SHIFT_REG0 = KEYNumberByte0           
        NVMDAT = $A5
        GoSub Make_EEPROMWrite_Sub       ' Write the enable code.
    Else         
        '=======================================================================
        ' READ THE PASSWORD FROM THE EEPROM:
        '=======================================================================
        ' The working registers are been written in the eeprom in case of power down.
        ' Read the 8 ASCII password and rebuilt the 4 Binary numbers for Random calculation.
        EEP_EEADR = EEP_ADDR_PASSWORD
        LOOPByte = 4
        Repeat
            Dec LOOPByte
            NVMCON1Bits_RD = 1
            DataByte = NVMDAT ^ KEYASCII        ' The ASCII bytes are XORED
            GoSub ASCII_to_BCD_Sub
            DataByte2 = WREG << 4
            Inc EEP_EEADR
            NVMCON1Bits_RD = 1
            DataByte = NVMDAT ^ KEYASCII        ' The ASCII bytes are XORED
            GoSub ASCII_to_BCD_Sub
            DataByte2 = DataByte2 | WREG           
            ShiftRegArray[LOOPByte] = DataByte2 ' SHIFT_REG3, SHIFT_REG2, SHIFT_REG1, SHIFT_REG0
            Inc EEP_EEADR
        Until LOOPByte = 0
    EndIf
    '---------------------------------------------------------------------------
    $else
    '---------------------------------------------------------------------------
    If NVMDAT <> $A5 Then                   ' Is it the first time?
        '=======================================================================
        ' LOAD THE ORIGINAL KEY FROM RAM ONLY: (Reset Password)
        '=======================================================================
        ' Load the original number of the project. (reset password)
        SHIFT_REG3 = KEYNumberByte3 
        SHIFT_REG2 = KEYNumberByte2  
        SHIFT_REG1 = KEYNumberByte1
        SHIFT_REG0 = KEYNumberByte0          
        NVMDAT = $A5
        GoSub Make_EEPROMWrite_Sub       ' Write the enable code.    
    EndIf
    '---------------------------------------------------------------------------
    $endif
    '===========================================================================
    ' RANDOM CODE STARTS:
    '===========================================================================
    TEMP_REG3A = SHIFT_REG3
    TEMP_REG3B = SHIFT_REG3
    NEW_REG0 = SHIFT_REG3
    TEMP_REG2A = SHIFT_REG2
    TEMP_REG2B = SHIFT_REG2
    STATUS = 0
    Asm
    Rlcf TEMP_REG2A,F
    Rlcf TEMP_REG3A,F
    Rlcf TEMP_REG2A,F
    Rlcf TEMP_REG3A,W
    Xorwf NEW_REG0,F
    Rrcf TEMP_REG3B,F
    Rrcf TEMP_REG2B,F
    Movf TEMP_REG2B,W
    Xorwf NEW_REG0,F
    Rrcf TEMP_REG3B,F
    Rrcf TEMP_REG2B,W
    Xorwf NEW_REG0,F 
    EndAsm
    SHIFT_REG3 = SHIFT_REG2 ^ SHIFT_REG0
    SHIFT_REG2 = SHIFT_REG1
    SHIFT_REG1 = SHIFT_REG0 ^ NEW_REG0
    SHIFT_REG0 = NEW_REG0
    '===========================================================================
    ' RANDOM CODE END:
    '===========================================================================
    ' SAVE THE PASSWORD IN THE EEPROM:
    '===========================================================================
    $ifndef _EnableToPrintCodes_
    ' Writing 1000 times to the eeprom is not allowed.
    ' Convert the 2 nibbles of the Binary Numbers to ASCII & Write the new ASCII PassWord to the eeprom.
    EEP_EEADR = EEP_ADDR_PASSWORD
    LOOPByte = 4
    Repeat
        Dec LOOPByte
        Byte_DSB = ShiftRegArray[LOOPByte]      ' SHIFT_REG3, SHIFT_REG2, SHIFT_REG1, SHIFT_REG0
        WREG = Byte_DSB >> 4                    ' Convert the high nibble.
        GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
        NVMDAT = WREG ^ KEYASCII                ' The ASCII bytes are XORED
        GoSub Make_EEPROMWrite_Sub                   ' Save to the eeprom.    
        Inc EEP_EEADR
        WREG = Byte_DSB                         ' Convert the low nibble.
        GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
        NVMDAT = WREG ^ KEYASCII                ' The ASCII bytes are XORED
        GoSub Make_EEPROMWrite_Sub                   ' Save to the eeprom.        
        Inc EEP_EEADR
    Until LOOPByte = 0    
    $endif
    Return
$endif
'*******************************************************************************
'* Name    : PrintNextPW_Sub                                                   *
'* Purpose : Print to the terminal the next Password                           *
'*******************************************************************************
$ifdef _EnableToPrintCodes_
PrintNextPW_Sub:
    ' Print the numbers from RAM of Passwords counter(in BCD).
    LOOPByte = 5
    Repeat
        Dec LOOPByte
        WREG = NumberOfPasswordArray[LOOPByte]
        GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub
        GoSub HRSOUT_Sub
    Until LOOPByte = 0
    WREG = " "
    GoSub HRSOUT_Sub    
    
    WREG = "P"
    GoSub HRSOUT_Sub
    WREG = "W"
    GoSub HRSOUT_Sub
    WREG = " "
    GoSub HRSOUT_Sub
    WREG = ":"
    GoSub HRSOUT_Sub

    ' Print the Passwords from RAM only.
    LOOPByte = 4
    Repeat
        Dec LOOPByte
        WREG = ShiftRegArray[LOOPByte]      ' SHIFT_REG3, SHIFT_REG2, SHIFT_REG1, SHIFT_REG0
        Conv1Byte_to_2ASCII(WREG)
    Until LOOPByte = 0  
    GoSub HRSOUT#CR_Sub
    Return
$endif
'*******************************************************************************
'* Name    : IncBCDCounters_Sub                                                *
'* Purpose : Increment the number of password (BCD counters)                   *
'* Instead to increment a Binary counter and transform the Binary to Decimal   *
'* it is more efficient to increment BCD counters directly.                    *
'*******************************************************************************
$ifdef _EnablePassWordMode0_
IncBCDCounters_Sub:
    $ifndef _EnableToPrintCodes_
    ' Writing 1000 times to the eeprom is not allowed.
    ' Read the eeprom & Load the PW counter with the values saved in the eeprom.
    EEP_EEADR = EEP_ADDR_NB_PASSWORD 
    LOOPByte = 0
    Repeat
        NVMCON1 = 0                                  ' To read the eeprom only
        NVMCON1Bits_RD = 1
        NumberOfPasswordArray[LOOPByte] = NVMDAT
        Inc LOOPByte
        Inc EEP_EEADR
    Until LOOPByte = 5   
    $endif
    
    ' Increment and save the counter to eeprom memory.
    ' The BCD counters are built with an array of 5 bytes => (00000 to 99999).
    $ifndef _EnableToPrintCodes_
    EEP_EEADR = EEP_ADDR_NB_PASSWORD
    $endif
    NumberBCDCounters = 5
    F_CarryFlag = 0
    FSR0 = VarPtr(NumberOfPasswordArray)
    Inc INDF0                       ' Increment the Low BCD counter
    Repeat
        If F_CarryFlag = 1 Then     ' If Carry set?
            Inc INDF0               ' Then increment again
            F_CarryFlag = 0         ' Done
        EndIf
        If INDF0 = 11 Then          ' 2 increments?
            F_CarryFlag = 1         ' Set Carry flag for next BCD counter
            INDF0 = 1               ' Then set one.
        EndIf
        If INDF0 = 10 Then          ' 1 increment?
            F_CarryFlag = 1         ' Set Carry flag for next BCD counter
            INDF0 = 0               ' Counter cleared.
        EndIf
        '-----------------------------------------------------------------------
        $ifndef _EnableToPrintCodes_
        ' Writing 1000 times to the eeprom is not allowed.
        ' Write a Byte to the eeprom
        NVMDAT = INDF0
        GoSub Make_EEPROMWrite_Sub
        Inc EEP_EEADR
        $endif
        '-----------------------------------------------------------------------
        Inc FSR0
        Dec NumberBCDCounters
    Until NumberBCDCounters = 0
    Return
$endif
'*******************************************************************************
'* Name    : LoadPassword1_Sub                                                 *
'* Purpose : Load the Key into an array to use it like a Password              *
'*******************************************************************************    
$ifdef _EnablePassWordMode1_
LoadPassword1_Sub:
    #ifdef watchdog_req
    Clrwdt
    #endif    
    PasswordMode1Array#0 = ":"              ' Load the first control.

    WREG = KEYNumberByte3 >> 4              ' Convert the high nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
    PasswordMode1Array#1 = WREG
    WREG = KEYNumberByte3                   ' Convert the low nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
    PasswordMode1Array#2 = WREG

    WREG = KEYNumberByte2 >> 4              ' Convert the high nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
    PasswordMode1Array#3 = WREG
    WREG = KEYNumberByte2                   ' Convert the low nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
    PasswordMode1Array#4 = WREG  

    WREG = KEYNumberByte1 >> 4              ' Convert the high nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
    PasswordMode1Array#5 = WREG
    WREG = KEYNumberByte1                   ' Convert the low nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
    PasswordMode1Array#6 = WREG

    WREG = KEYNumberByte0 >> 4              ' Convert the high nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
    PasswordMode1Array#7 = WREG
    WREG = KEYNumberByte0                   ' Convert the low nibble.
    GoSub CONVERT_LOWNIBBLE_to_ASCII_Sub    ' In WREG
    PasswordMode1Array#8 = WREG
    Return
$endif  
'###############################################################################
'*******************************************************************************
CONFIG19: '* (19) DEVELOPER:                                                   *
'* Name    : ADDRESS OF THE PASSWORD SYSTEM IN THE EEPROM: (EVEN NUMBERS) => 16*
'*******************************************************************************
$ifdef _EnablePassWordMode0_
    Org EEP_ADDRESS + EEP_ADDR_PASSWORD_SYSTEM  ' _eeprom - 16 (15728640 + 256 - 16)
    CData 0,0,0,0,0,PW_CONTROL                  ' Head _eeprom - 16  , Number of Password (Binary), Password Control (Binary)
    CData 255,65,65,65,65,65,65,65,65,255       ' Head _eeprom - 10  , Enable writing (Binary) ; Password (ASCII) (AAAAAAAA)
$endif
'*******************************************************************************
CONFIG19A: '* (19A) DEVELOPER:                                                 *
'* Name    : USER code Origins (Version 4.0)                                   *
'* Purpose : Place the origins For the user Code                               *
'* Determined by the "Declare PROTON_START_ADDRESS = XXXX" in user code        *
'*******************************************************************************
    Org ORG_Proton_Start    ' Address: USER_PROTON_START_ADDRESS:  Redirected to Proton#Code#Start
ProtonCodeStart:
    ' To protect the User Memory Code against overlap.
    While 1 = 1 : Wend      ' Error 118: Overwiting previous address contents $0800.

    Org ORG_ISR_High        ' New User Hardware Interrupt
ISR_High:

    Org ORG_ISR_Low         ' New User Low Interrupt
ISR_Low:
    End               
'===============================================================================
'###############################################################################
'*******************************************************************************
CONFIG20: '* (20) DEVELOPER:                                                   *
'* Name    : ERASE the IDLOCS positions At compile time, And when this firmware*
'*           into the PIC by a programmer.                                     *
'*******************************************************************************
$ifdef _EnableWriteIDLOCS_
Asm
$ifdef _EraseIDlocs_
__IDLOCS   IDLOC0,0xFF
__IDLOCS   IDLOC1,0xFF
__IDLOCS   IDLOC2,0xFF
__IDLOCS   IDLOC3,0xFF
__IDLOCS   IDLOC4,0xFF
__IDLOCS   IDLOC5,0xFF
__IDLOCS   IDLOC6,0xFF
__IDLOCS   IDLOC7,0xFF
$else
__IDLOCS   IDLOC0,_IDLOC0_
__IDLOCS   IDLOC1,_IDLOC1_
__IDLOCS   IDLOC2,_IDLOC2_
__IDLOCS   IDLOC3,_IDLOC3_
__IDLOCS   IDLOC4,_IDLOC4_
__IDLOCS   IDLOC5,_IDLOC5_
__IDLOCS   IDLOC6,_IDLOC6_
__IDLOCS   IDLOC7,_IDLOC7_
$endif
EndAsm
$endif
'###############################################################################
'######## END OF FILE ##########################################################
'###############################################################################
	
